import ast
import math
import os
import logging

from qtpy.QtWidgets import (QWidget, QStyle, QStyleOption)
from qtpy.QtGui import (QColor, QPainter, QBrush, QPen, QPolygon, QPolygonF, QPixmap,
                            QMovie)
from qtpy.QtCore import Property, Qt, QPoint, QPointF, QSize, Slot, QTimer, QRectF
from qtpy.QtDesigner import QDesignerFormWindowInterface
from .base import PyDMWidget
from ..utilities import is_qt_designer, find_file

logger = logging.getLogger(__name__)


def deg_to_qt(deg):
    """
    Converts from degrees to QT degrees.
    16 deg = 1 QTdeg

    Parameters
    ----------
    deg : float
        The value to convert.

    Returns
    -------
    float
        The value converted.
    """
    # Angles for Qt are in units of 1/16 of a degree
    return deg * 16


def qt_to_deg(deg):
    """
    Converts from QT degrees to degrees.
    16 deg = 1 QTdeg

    Parameters
    ----------
    deg : float
        The value to convert.

    Returns
    -------
    float
        The value converted.
    """
    # Angles for Qt are in units of 1/16 of a degree
    return deg / 16.0


class PyDMDrawing(QWidget, PyDMWidget):
    """
    Base class to be used for all PyDM Drawing Widgets.
    This class inherits from QWidget and PyDMWidget.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the Label
    init_channel : str, optional
        The channel to be used by the widget.
    """
    def __init__(self, parent=None, init_channel=None):
        self._rotation = 0.0
        self._brush = QBrush(Qt.SolidPattern)
        self._original_brush = self._brush
        self._painter = QPainter()
        self._pen = QPen(Qt.NoPen)
        self._pen_style = Qt.NoPen
        self._pen_cap_style = Qt.SquareCap
        self._pen_join_style = Qt.MiterJoin
        self._pen_width = 0
        self._pen_color = QColor(0, 0, 0)
        self._pen.setCapStyle(self._pen_cap_style)
        self._pen.setJoinStyle(self._pen_join_style)
        self._original_pen_style = self._pen_style
        self._original_pen_color = self._pen_color
        QWidget.__init__(self, parent)
        PyDMWidget.__init__(self, init_channel=init_channel)
        self.alarmSensitiveBorder = False

    def sizeHint(self):
        return QSize(100, 100)

    def paintEvent(self, _):
        """
        Paint events are sent to widgets that need to update themselves,
        for instance when part of a widget is exposed because a covering
        widget was moved.

        At PyDMDrawing this method handles the alarm painting with parameters
        from the stylesheet, configures the brush, pen and calls ```draw_item```
        so the specifics can be performed for each of the drawing classes.

        Parameters
        ----------
        event : QPaintEvent
        """
        painter = QPainter(self)
        opt = QStyleOption()
        opt.initFrom(self)
        self.style().drawPrimitive(QStyle.PE_Widget, opt, painter, self)
        painter.setRenderHint(QPainter.Antialiasing)

        painter.setBrush(self._brush)
        painter.setPen(self._pen)

        self.draw_item(painter)

    def draw_item(self, painter):
        """
        The classes inheriting from PyDMDrawing must overwrite this method.
        This method translate the painter to the center point given by
        ```get_center``` and rotate the canvas by the given amount of
        degrees.
        """
        xc, yc = self.get_center()
        painter.translate(xc, yc)
        painter.rotate(-self._rotation)

    def get_center(self):
        """
        Simple calculation of the canvas' center point.

        Returns
        -------
        x, y : float
            Tuple with X and Y coordinates of the center.
        """
        return self.width() * 0.5, self.height() * 0.5

    def get_bounds(self, maxsize=False, force_no_pen=False):
        """
        Returns a tuple containing the useful area for the drawing.

        Parameters
        ----------
        maxsize : bool, default is False
            If True, width and height information are based on the
            maximum inner rectangle dimensions given by ```get_inner_max```,
            otherwise width and height will receive the widget size.

        force_no_pen : bool, default is False
            If True the pen width will not be considered when calculating
            the bounds.

        Returns
        -------
        x, y, w, h : tuple
            Tuple with X and Y coordinates followed by the maximum width
            and height.
        """
        w, h = self.width(), self.height()

        if maxsize:
            w, h = self.get_inner_max()

        xc, yc = w * 0.5, h * 0.5

        if self.has_border() and not force_no_pen:
            w = max(0, w - 2 * self._pen_width)
            h = max(0, h - 2 * self._pen_width)
            x = max(0, self._pen_width)
            y = max(0, self._pen_width)
        else:
            x = 0
            y = 0
        return x - xc, y - yc, w, h

    def has_border(self):
        """
        Check whether or not the drawing have a border based on the
        Pen Style and Pen width.

        Returns
        -------
        bool
            True if the drawing has a border, False otherwise.
        """
        if self._pen.style() != Qt.NoPen and self._pen_width > 0:
            return True
        else:
            return False

    def is_square(self):
        """
        Check if the widget has the same width and height values.

        Returns
        -------
        bool
            True in case the widget has a square shape, False otherwise.
        """
        return self.height() == self.width()

    def get_inner_max(self):
        """
        Calculates the largest inner rectangle in a rotated rectangle.
        This implementation was based on https://stackoverflow.com/a/18402507

        Returns
        -------
        w, h : tuple
            The width and height of the largest rectangle.
        """
        # Based on https://stackoverflow.com/a/18402507
        w0 = 0
        h0 = 0
        angle = math.radians(self._rotation)
        origWidth = self.width()
        origHeight = self.height()

        if origWidth == 0:
            logger.error("Invalid width. The value must be greater than {0}".format(origWidth))
            return

        if origHeight == 0:
            logger.error("Invalid height. The value must be greater than {0}".format(origHeight))
            return

        if (origWidth <= origHeight):
            w0 = origWidth
            h0 = origHeight
        else:
            w0 = origHeight
            h0 = origWidth
        # Angle normalization in range [-PI..PI)
        ang = angle - math.floor((angle + math.pi) / (2 * math.pi)) * 2 * math.pi
        ang = math.fabs(ang)
        if ang > math.pi / 2:
            ang = math.pi - ang
        c = w0 / (h0 * math.sin(ang) + w0 * math.cos(ang))
        w = 0
        h = 0
        if (origWidth <= origHeight):
            w = w0 * c
            h = h0 * c
        else:
            w = h0 * c
            h = w0 * c
        return w, h

    @Property(QBrush)
    def brush(self):
        """
        PyQT Property for the brush object to be used when coloring the
        drawing

        Returns
        -------
        QBrush
        """
        return self._brush

    @brush.setter
    def brush(self, new_brush):
        """
        PyQT Property for the brush object to be used when coloring the
        drawing

        Parameters
        ----------
        new_brush : QBrush
        """
        if new_brush != self._brush:
            if self._alarm_state == PyDMWidget.ALARM_NONE:
                self._original_brush = new_brush
            self._brush = new_brush
            self.update()

    @Property(Qt.PenStyle)
    def penStyle(self):
        """
        PyQT Property for the pen style to be used when drawing the border

        Returns
        -------
        int
            Index at Qt.PenStyle enum
        """
        return self._pen_style

    @penStyle.setter
    def penStyle(self, new_style):
        """
        PyQT Property for the pen style to be used when drawing the border

        Parameters
        ----------
        new_style : int
            Index at Qt.PenStyle enum
        """
        if self._alarm_state == PyDMWidget.ALARM_NONE:
            self._original_pen_style = new_style
        if new_style != self._pen_style:
            self._pen_style = new_style
            self._pen.setStyle(new_style)
            self.update()

    @Property(Qt.PenCapStyle)
    def penCapStyle(self):
        """
        PyQT Property for the pen cap to be used when drawing the border

        Returns
        -------
        int
            Index at Qt.PenCapStyle enum
        """
        return self._pen_cap_style

    @penCapStyle.setter
    def penCapStyle(self, new_style):
        """
        PyQT Property for the pen cap style to be used when drawing the border

        Parameters
        ----------
        new_style : int
            Index at Qt.PenStyle enum
        """
        if new_style != self._pen_cap_style:
            self._pen_cap_style = new_style
            self._pen.setCapStyle(new_style)
            self.update()

    @Property(Qt.PenJoinStyle)
    def penJoinStyle(self):
        """
        PyQT Property for the pen join style to be used when drawing the border

        Returns
        -------
        int
            Index at Qt.PenJoinStyle enum
        """
        return self._pen_join_style

    @penJoinStyle.setter
    def penJoinStyle(self, new_style):
        """
        PyQT Property for the pen join style to be used when drawing the border

        Parameters
        ----------
        new_style : int
            Index at Qt.PenStyle enum
        """
        if new_style != self._pen_join_style:
            self._pen_join_style = new_style
            self._pen.setJoinStyle(new_style)
            self.update()

    @Property(QColor)
    def penColor(self):
        """
        PyQT Property for the pen color to be used when drawing the border

        Returns
        -------
        QColor
        """
        return self._pen_color

    @penColor.setter
    def penColor(self, new_color):
        """
        PyQT Property for the pen color to be used when drawing the border

        Parameters
        ----------
        new_color : QColor
        """
        if self._alarm_state == PyDMWidget.ALARM_NONE:
            self._original_pen_color = new_color

        if new_color != self._pen_color:
            self._pen_color = new_color
            self._pen.setColor(new_color)
            self.update()

    @Property(float)
    def penWidth(self):
        """
        PyQT Property for the pen width to be used when drawing the border

        Returns
        -------
        float
        """
        return self._pen_width

    @penWidth.setter
    def penWidth(self, new_width):
        """
        PyQT Property for the pen width to be used when drawing the border

        Parameters
        ----------
        new_width : float
        """
        if new_width < 0:
            return
        if new_width != self._pen_width:
            self._pen_width = new_width
            self._pen.setWidthF(float(self._pen_width))
            self.update()

    @Property(float)
    def rotation(self):
        """
        PyQT Property for the counter-clockwise rotation in degrees
        to be applied to the drawing.

        Returns
        -------
        float
        """
        return self._rotation

    @rotation.setter
    def rotation(self, new_angle):
        """
        PyQT Property for the counter-clockwise rotation in degrees
        to be applied to the drawing.

        Parameters
        ----------
        new_angle : float
        """
        if new_angle != self._rotation:
            self._rotation = new_angle
            self.update()

    def alarm_severity_changed(self, new_alarm_severity):
        PyDMWidget.alarm_severity_changed(self, new_alarm_severity)
        if new_alarm_severity == PyDMWidget.ALARM_NONE:
            if self._original_brush is not None:
                self.brush = self._original_brush
            if self._original_pen_color is not None:
                self.penColor = self._original_pen_color
            if self._original_pen_style is not None:
                self.penStyle = self._original_pen_style


class PyDMDrawingLine(PyDMDrawing):
    """
    A widget with a line drawn in it.
    This class inherits from PyDMDrawing.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the Label
    init_channel : str, optional
        The channel to be used by the widget.
    """

    def __init__(self, parent=None, init_channel=None):
        super(PyDMDrawingLine, self).__init__(parent, init_channel)
        self._arrow_end_point_selection = False
        self._arrow_start_point_selection = False
        self._arrow_mid_point_selection = False
        self._mid_point_arrow_flipped = False
        self.rotation = 0
        self.penStyle = Qt.SolidLine
        self.penWidth = 1

    @staticmethod
    def _arrow_points(startpoint, endpoint, height, width):
        """
        Returns the three points needed to make a triangle with .drawPolygon
        """
        diff_x = startpoint.x() - endpoint.x()
        diff_y = startpoint.y() - endpoint.y()

        length = math.sqrt(diff_x ** 2 + diff_y ** 2)

        norm_x = diff_x / length
        norm_y = diff_y / length

        perp_x = -norm_y
        perp_y = norm_x

        left_x = endpoint.x() + height * norm_x + width * perp_x
        left_y = endpoint.y() + height * norm_y + width * perp_y
        right_x = endpoint.x() + height * norm_x - width * perp_x
        right_y = endpoint.y() + height * norm_y - width * perp_y

        left = QPointF(left_x, left_y)
        right = QPointF(right_x, right_y)

        return QPolygonF([left, endpoint, right])

    def draw_item(self, painter):
        """
        Draws the line after setting up the canvas with a call to
        ```PyDMDrawing.draw_item```.
        """
        super(PyDMDrawingLine, self).draw_item(painter)
        x, y, w, h = self.get_bounds()

        # Figure out how long to make the line to touch the bounding box
        # Length varies depending on rotation

        # Find the quadrant 1 angle equivalent
        angle = self._rotation % 360
        if 90 < angle <= 180:
            angle = 180 - angle
        elif 180 < angle <= 270:
            angle = angle - 180
        elif 270 < angle <= 360:
            angle = 360 - angle
        angle_rad = math.radians(angle)

        # Find the angle of the rop right corner of the bounding box
        try:
            critical_angle = math.atan(h/w)
        except ZeroDivisionError:
            critical_angle = math.pi / 2

        # Pick a length based on which side we intersect with
        if angle_rad > critical_angle:
            try:
                length = h / math.sin(angle_rad)
            except ZeroDivisionError:
                length = w
        else:
            try:
                length = w / math.cos(angle_rad)
            except ZeroDivisionError:
                length = h

        # Define endpoints potentially outside the bounding box
        # Will land on the bounding box after rotation
        midpoint = x + w/2
        start_point = QPointF(midpoint - length/2, 0)
        end_point = QPointF(midpoint + length/2, 0)
        mid_point = QPointF(midpoint, 0)

        # Draw the line
        painter.drawLine(start_point, end_point)

        # Draw the arrows
        if self._arrow_end_point_selection:
            points = self._arrow_points(start_point, end_point, 6, 6)
            painter.drawPolygon(points)

        if self._arrow_start_point_selection:
            points = self._arrow_points(end_point, start_point, 6, 6)
            painter.drawPolygon(points)

        if self._arrow_mid_point_selection:
            if self._mid_point_arrow_flipped:
                points = self._arrow_points(start_point, mid_point, 6, 6)
            else:
                points = self._arrow_points(end_point, mid_point, 6, 6)
            painter.drawPolygon(points)

    @Property(bool)
    def arrowEndPoint(self):
        """
        If True, an arrow will be drawn at the end of the line.

        Returns
        -------
        bool
        """
        return self._arrow_end_point_selection

    @arrowEndPoint.setter
    def arrowEndPoint(self, new_selection):
        """
        If True, an arrow will be drawn at the end of the line.

        Parameters
        -------
        new_selection : bool
        """
        if self._arrow_end_point_selection != new_selection:
            self._arrow_end_point_selection = new_selection
            self.update()

    @Property(bool)
    def arrowStartPoint(self):
        """
        If True, an arrow will be drawn at the start of the line.

        Returns
        -------
        bool
        """
        return self._arrow_start_point_selection

    @arrowStartPoint.setter
    def arrowStartPoint(self, new_selection):
        """
        If True, an arrow will be drawn at the start of the line.

        Parameters
        -------
        new_selection : bool
        """
        if self._arrow_start_point_selection != new_selection:
            self._arrow_start_point_selection = new_selection
            self.update()

    @Property(bool)
    def arrowMidPoint(self):
        """
        If True, an arrow will be drawn at the midpoint of the line.
        Returns
        -------
        bool
        """
        return self._arrow_mid_point_selection

    @arrowMidPoint.setter
    def arrowMidPoint(self, new_selection):
        """
        If True, an arrow will be drawn at the midpoint of the line.
        Parameters
        -------
        new_selection : bool
        """
        if self._arrow_mid_point_selection != new_selection:
            self._arrow_mid_point_selection = new_selection
            self.update()

    @Property(bool)
    def flipMidPointArrow(self):
        """
        Flips the direction of the midpoint arrow.

        Returns
        -------
        bool
        """
        return self._mid_point_arrow_flipped

    @flipMidPointArrow.setter
    def flipMidPointArrow(self, new_selection):
        """
        Flips the direction of the midpoint arrow.
        
        Parameters
        -------
        new_selection : bool
        """
        if self._mid_point_arrow_flipped != new_selection:
            self._mid_point_arrow_flipped = new_selection
            self.update()

class PyDMDrawingImage(PyDMDrawing):
    """
    Renders an image given by the ``filename`` property.
    This class inherits from PyDMDrawing.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the Label
    init_channel : str, optional
        The channel to be used by the widget.

    Attributes
    ----------
    null_color : Qt.Color
        QColor to fill the image if the filename is not found.
    """
    null_color = Qt.gray

    def __init__(self, parent=None, init_channel=None, filename=""):
        super(PyDMDrawingImage, self).__init__(parent, init_channel)
        hint = super(PyDMDrawingImage, self).sizeHint()
        self._pixmap = QPixmap(hint)
        self._pixmap.fill(self.null_color)
        self._aspect_ratio_mode = Qt.KeepAspectRatio
        self._movie = None
        self._file = None
        # Make sure we don't set a non-existant file
        if filename:
            self.filename = filename
        # But we always have an internal value to reference
        else:
            self._file = filename
        if is_qt_designer():  # pragma: no cover
            designer_window = self.get_designer_window()
            if designer_window is not None:
                designer_window.fileNameChanged.connect(self.designer_form_saved)
                QTimer.singleShot(200, self.reload_image)

    def get_designer_window(self):  # pragma: no cover
        # Internal function to find the designer window that owns this widget.
        p = self.parent()
        while p is not None:
            if isinstance(p, QDesignerFormWindowInterface):
                return p
            p = p.parent()
        return None

    @Slot(str)
    def designer_form_saved(self, filename):  # pragma: no cover
        self.filename = self._file

    def reload_image(self):
        self.filename = self._file

    @Property(str)
    def filename(self):
        """
        The filename of the image to be displayed.
        This can be an absolute or relative path to the display file.

        Returns
        -------
        str
            The filename configured.
        """
        return self._file

    @filename.setter
    def filename(self, new_file):
        """
        The filename of the image to be displayed.

        This file can be either relative to the ``.ui`` file or absolute. If
        the path does not exist, a shape of ``.null_color`` will be displayed
        instead.

        Parameters
        -------
        new_file : str
            The filename to be used
        """
        # Expand user (~ or ~user) and environment variables.
        pixmap = None
        self._file = new_file
        abs_path = os.path.expanduser(os.path.expandvars(self._file))
        # Find the absolute path relative to UI
        if not os.path.isabs(abs_path):
            parent_display = self.find_parent_display()
            base_path = None
            if parent_display:
                base_path = os.path.dirname(parent_display.loaded_file())
            abs_path = find_file(abs_path, base_path=base_path)
            if not abs_path:
                logger.error("Unable to find full filepath for %s", self._file)
                return

        # Check that the path exists
        if os.path.isfile(abs_path):
            if self._movie is not None:
                self._movie.stop()
                self._movie.deleteLater()
                self._movie = None
            if not abs_path.endswith(".gif"):
                pixmap = QPixmap(abs_path)
            else:
                self._movie = QMovie(abs_path, parent=self)
                self._movie.setCacheMode(QMovie.CacheAll)
                self._movie.frameChanged.connect(self.movie_frame_changed)
                if self._movie.frameCount() > 1:
                    self._movie.finished.connect(self.movie_finished)
                self._movie.start()

        # Return a blank image if we don't have a valid path
        else:
            # Warn the user loudly if their file does not exist, but avoid
            # doing this in Designer as this spams the user as they are typing
            if not is_qt_designer():  # pragma: no cover
                logger.error("Image file  %r does not exist", abs_path)
            pixmap = QPixmap(self.sizeHint())
            pixmap.fill(self.null_color)
        # Update the display
        if pixmap is not None:
            self._pixmap = pixmap
            self.update()

    def sizeHint(self):
        if self._pixmap.size().isEmpty():
            return super(PyDMDrawingImage, self).sizeHint()
        return self._pixmap.size()

    @Property(Qt.AspectRatioMode)
    def aspectRatioMode(self):
        """
        PyQT Property for aspect ratio mode to be used when rendering
        the image

        Returns
        -------
        int
            Index at Qt.AspectRatioMode enum
        """
        return self._aspect_ratio_mode

    @aspectRatioMode.setter
    def aspectRatioMode(self, new_mode):
        """
        PyQT Property for aspect ratio mode to be used when rendering
        the image

        Parameters
        ----------
        new_mode : int
            Index at Qt.AspectRatioMode enum
        """
        if new_mode != self._aspect_ratio_mode:
            self._aspect_ratio_mode = new_mode
            self.update()

    def draw_item(self, painter):
        """
        Draws the image after setting up the canvas with a call to
        ```PyDMDrawing.draw_item```.
        """
        super(PyDMDrawingImage, self).draw_item(painter)
        x, y, w, h = self.get_bounds(maxsize=True, force_no_pen=True)
        if not isinstance(self._pixmap, QMovie):
            _scaled = self._pixmap.scaled(int(w), int(h), self._aspect_ratio_mode,
                                          Qt.SmoothTransformation)
            # Make sure the image is centered if smaller than the widget itself
            if w > _scaled.width():
                logger.debug("Centering image horizontally ...")
                x += (w-_scaled.width())/2
            if h > _scaled.height():
                logger.debug("Centering image vertically ...")
                y += (h - _scaled.height())/2
            painter.drawPixmap(QPointF(x, y), _scaled)

    def movie_frame_changed(self, frame_no):
        """
        Callback executed when a new frame is available at the QMovie.

        Parameters
        ----------
        frame_no : int
            The new frame index

        Returns
        -------
        None

        """
        if self._movie is None:
            return
        curr_pixmap = self._movie.currentPixmap()
        self._pixmap = curr_pixmap
        self.update()

    def movie_finished(self):
        """
        Callback executed when the movie is finished.

        Returns
        -------
        None
        """
        if self._movie is None:
            return

        self._movie.start()


class PyDMDrawingRectangle(PyDMDrawing):
    """
    A widget with a rectangle drawn in it.
    This class inherits from PyDMDrawing.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the Label
    init_channel : str, optional
        The channel to be used by the widget.
    """
    def __init__(self, parent=None, init_channel=None):
        super(PyDMDrawingRectangle, self).__init__(parent, init_channel)

    def draw_item(self, painter):
        """
        Draws the rectangle after setting up the canvas with a call to
        ```PyDMDrawing.draw_item```.
        """
        super(PyDMDrawingRectangle, self).draw_item(painter)
        x, y, w, h = self.get_bounds(maxsize=True)
        painter.drawRect(QRectF(x, y, w, h))


class PyDMDrawingTriangle(PyDMDrawing):
    """
    A widget with a triangle drawn in it.
    This class inherits from PyDMDrawing.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the Label
    init_channel : str, optional
        The channel to be used by the widget.
    """
    def __init__(self, parent=None, init_channel=None):
        super(PyDMDrawingTriangle, self).__init__(parent, init_channel)

    def _calculate_drawing_points(self, x, y, w, h):
        return [
            QPointF(x, h / 2.0),
            QPointF(x, y),
            QPointF(w / 2.0, y)
        ]

    def draw_item(self, painter):
        """
        Draws the triangle after setting up the canvas with a call to
        ```PyDMDrawing.draw_item```.
        """
        super(PyDMDrawingTriangle, self).draw_item(painter)
        x, y, w, h = self.get_bounds(maxsize=True)
        points = self._calculate_drawing_points(x, y, w, h)

        painter.drawPolygon(QPolygonF(points))


class PyDMDrawingEllipse(PyDMDrawing):
    """
    A widget with an ellipse drawn in it.
    This class inherits from PyDMDrawing.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the Label
    init_channel : str, optional
        The channel to be used by the widget.
    """
    def __init__(self, parent=None, init_channel=None):
        super(PyDMDrawingEllipse, self).__init__(parent, init_channel)

    def draw_item(self, painter):
        """
        Draws the ellipse after setting up the canvas with a call to
        ```PyDMDrawing.draw_item```.
        """
        super(PyDMDrawingEllipse, self).draw_item(painter)
        maxsize = not self.is_square()
        _, _, w, h = self.get_bounds(maxsize=maxsize)
        painter.drawEllipse(QPoint(0, 0), w / 2.0, h / 2.0)


class PyDMDrawingCircle(PyDMDrawing):
    """
    A widget with a circle drawn in it.
    This class inherits from PyDMDrawing.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the Label
    init_channel : str, optional
        The channel to be used by the widget.
    """
    def __init__(self, parent=None, init_channel=None):
        super(PyDMDrawingCircle, self).__init__(parent, init_channel)

    def _calculate_radius(self, width, height):
        return min(width, height) / 2.0

    def draw_item(self, painter):
        """
        Draws the circle after setting up the canvas with a call to
        ```PyDMDrawing.draw_item```.
        """
        super(PyDMDrawingCircle, self).draw_item(painter)
        _, _, w, h = self.get_bounds()
        r = self._calculate_radius(w, h)
        painter.drawEllipse(QPoint(0, 0), r, r)


class PyDMDrawingArc(PyDMDrawing):
    """
    A widget with an arc drawn in it.
    This class inherits from PyDMDrawing.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the Label
    init_channel : str, optional
        The channel to be used by the widget.
    """
    def __init__(self, parent=None, init_channel=None):
        super(PyDMDrawingArc, self).__init__(parent, init_channel)
        self.penStyle = Qt.SolidLine
        self.penWidth = 1.0
        self._start_angle = 0
        self._span_angle = deg_to_qt(90)

    @Property(float)
    def startAngle(self):
        """
        PyQT Property for the start angle in degrees

        Returns
        -------
        float
            Angle in degrees
        """
        return qt_to_deg(self._start_angle)

    @startAngle.setter
    def startAngle(self, new_angle):
        """
        PyQT Property for the start angle in degrees

        Parameters
        ----------
        new_angle : float
            Angle in degrees
        """
        if deg_to_qt(new_angle) != self._start_angle:
            self._start_angle = deg_to_qt(new_angle)
            self.update()

    @Property(float)
    def spanAngle(self):
        """
        PyQT Property for the span angle in degrees

        Returns
        -------
        float
            Angle in degrees
        """
        return qt_to_deg(self._span_angle)

    @spanAngle.setter
    def spanAngle(self, new_angle):
        """
        PyQT Property for the span angle in degrees

        Parameters
        ----------
        new_angle : float
            Angle in degrees
        """
        if deg_to_qt(new_angle) != self._span_angle:
            self._span_angle = deg_to_qt(new_angle)
            self.update()

    def draw_item(self, painter):
        """
        Draws the arc after setting up the canvas with a call to
        ```PyDMDrawing.draw_item```.
        """
        super(PyDMDrawingArc, self).draw_item(painter)
        maxsize = not self.is_square()
        x, y, w, h = self.get_bounds(maxsize=maxsize)
        painter.drawArc(QRectF(x, y, w, h), int(self._start_angle), int(self._span_angle))


class PyDMDrawingPie(PyDMDrawingArc):
    """
    A widget with a pie drawn in it.
    This class inherits from PyDMDrawing.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the Label
    init_channel : str, optional
        The channel to be used by the widget.
    """
    def __init__(self, parent=None, init_channel=None):
        super(PyDMDrawingPie, self).__init__(parent, init_channel)

    def draw_item(self, painter):
        """
        Draws the pie after setting up the canvas with a call to
        ```PyDMDrawing.draw_item```.
        """
        super(PyDMDrawingPie, self).draw_item(painter)
        maxsize = not self.is_square()
        x, y, w, h = self.get_bounds(maxsize=maxsize)
        painter.drawPie(QRectF(x, y, w, h), int(self._start_angle), int(self._span_angle))


class PyDMDrawingChord(PyDMDrawingArc):
    """
    A widget with a chord drawn in it.
    This class inherits from PyDMDrawing.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the Label
    init_channel : str, optional
        The channel to be used by the widget.
    """
    def __init__(self, parent=None, init_channel=None):
        super(PyDMDrawingChord, self).__init__(parent, init_channel)

    def draw_item(self, painter):
        """
        Draws the chord after setting up the canvas with a call to
        ```PyDMDrawing.draw_item```.
        """
        super(PyDMDrawingChord, self).draw_item(painter)
        maxsize = not self.is_square()
        x, y, w, h = self.get_bounds(maxsize=maxsize)
        painter.drawChord(QRectF(x, y, w, h), int(self._start_angle), int(self._span_angle))


class PyDMDrawingPolygon(PyDMDrawing):
    """
    A widget with a polygon drawn in it.
    This class inherits from PyDMDrawing.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the Label
    init_channel : str, optional
        The channel to be used by the widget.
    """
    def __init__(self, parent=None, init_channel=None):
        super(PyDMDrawingPolygon, self).__init__(parent, init_channel)
        self._num_points = 3

    @Property(int)
    def numberOfPoints(self):
        """
        PyQT Property for the number of points

        Returns
        -------
        int
            Number of Points
        """
        return self._num_points

    @numberOfPoints.setter
    def numberOfPoints(self, points):
        if points >= 3 and points != self._num_points:
            self._num_points = points
            self.update()

    def _calculate_drawing_points(self, x, y, w, h):
        #(x + r*cos(theta), y + r*sin(theta))
        r = min(w, h)/2.0
        deg_step = 360.0/self._num_points

        points = []
        for i in range(self._num_points):
            xp = r * math.cos(math.radians(deg_step * i))
            yp = r * math.sin(math.radians(deg_step * i))
            points.append(QPointF(xp, yp))

        return points

    def draw_item(self, painter):
        """
        Draws the Polygon after setting up the canvas with a call to
        ```PyDMDrawing.draw_item```.
        """
        super(PyDMDrawingPolygon, self).draw_item(painter)
        maxsize = not self.is_square()
        x, y, w, h = self.get_bounds(maxsize=not self.is_square())
        poly = self._calculate_drawing_points(x, y, w, h)
        painter.drawPolygon(QPolygonF(poly))


class PyDMDrawingPolyline(PyDMDrawing):
    """
    A widget with a multi-segment, piecewise-linear line drawn in it.
    This class inherits from PyDMDrawing.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the Label
    init_channel : str, optional
        The channel to be used by the widget.
    """
    def __init__(self, parent=None, init_channel=None):
        super(PyDMDrawingPolyline, self).__init__(parent, init_channel)
        self._arrow_end_point_selection = False
        self._arrow_start_point_selection = False
        self._arrow_mid_point_selection = False
        self._arrow_mid_point_flipped = False
        self.penStyle = Qt.SolidLine
        self.penWidth = 1
        self._points = []

    def draw_item(self, painter):
        """
        Draws the segmented line after setting up the canvas with a call to
        ``PyDMDrawing.draw_item``.
        """
        super(PyDMDrawingPolyline, self).draw_item(painter)
        x, y, w, h = self.get_bounds()

        def p2d(pt):
            "convert point to drawing coordinates"
            # drawing coordinates are centered: (0,0) is in center
            # our points are absolute: (0,0) is upper-left corner
            if isinstance(pt, str):
                # 2022-05-11: needed for backwards compatibility support
                # PyDM releases up to v1.15.1
                # adl2pydm tags up to 0.0.2
                pt = tuple(map(int, pt.split(",")))
            u, v = pt
            return QPointF(u+x, v+y)

        if len(self._points) > 1:
            for i, p1 in enumerate(self._points[:-1]):
                painter.drawLine(p2d(p1), p2d(self._points[i+1]))
                if self._arrow_mid_point_selection:
                    point1 = p2d(p1)
                    point2 = p2d(self._points[i+1])
                    if self._arrow_mid_point_flipped:
                        point1, point2 = point2, point1 #swap values

                    # arrow points at midpoint of line
                    midpoint_x = (point1.x() + point2.x()) / 2
                    midpoint_y = (point1.y() + point2.y()) / 2
                    midpoint = QPointF(midpoint_x, midpoint_y)
                    points = PyDMDrawingLine._arrow_points(point1, midpoint, 6, 6) # 6 = arbitrary arrow size
                    painter.drawPolygon(points)

        # Draw the arrows
        if self._arrow_end_point_selection and (len(self._points[1]) >= 2):
            points = PyDMDrawingLine._arrow_points(p2d(self._points[1]), p2d(self._points[0]), 6, 6)
            painter.drawPolygon(points)

        if self._arrow_start_point_selection and (len(self._points[1]) >= 2):
            points = PyDMDrawingLine._arrow_points(p2d(self._points[len(self._points)-2]), p2d(self._points[len(self._points)-1]), 6, 6)
            painter.drawPolygon(points)

    def getPoints(self):
        """Convert internal points representation for use as QStringList."""
        points = [
            f"{pt[0]}, {pt[1]}"
            for pt in self._points
        ]
        return points

    def _validator(self, value):
        """
        ensure that `value` has correct form

        Parameters
        ----------
        value : [ordered pairs]
            List of strings representing ordered pairs
            of integer coordinates.  Each ordered pair
            is a tuple or list.

        Returns
        ----------
        verified : [ordered pairs]
            List of `tuple(number, number)`.

        """
        def isfloat(value):
            if isinstance(value, str):
                value = value.strip()
            try:
                float(value)
                return True
            except:
                return False

        def validate_point(i, point):
            """Ignore (instead of fail on) any of these pathologies."""
            if isinstance(point, str):
                try:
                    point = ast.literal_eval(point)
                except SyntaxError:
                    logger.error("point %d must be two numbers, comma-separated, received '%s'", i, pt)
                    return
            if not isinstance(point, (list, tuple)) or len(point) != 2:
                logger.error("point %d must be two numbers, comma-separated, received '%s'", i, pt)
                return
            try:
                point = list(map(float, point))  # ensure all values are float
            except ValueError:
                logger.error("point %d content must be numeric, received '%s'", i, pt)
                return

            return point

        verified = []
        for i, pt in enumerate(value, start=1):
            point = validate_point(i, pt)
            if point is not None:
                verified.append(point)

        return verified

    def setPoints(self, value):
        verified = self._validator(value)
        if verified is not None:
            if len(verified) < 2:
                logger.error("Must have two or more points")
                return

            self._points = verified
            self.update()

    def resetPoints(self):
        self._points = []
        self.update()

    @Property(bool)
    def arrowEndPoint(self):
        """
        If True, an arrow will be drawn at the start of the last polyline segment.

        Returns
        -------
        bool
        """
        return self._arrow_end_point_selection

    @arrowEndPoint.setter
    def arrowEndPoint(self, new_selection):
        """
        If True, an arrow will be drawn at the start of the last polyline segment.

        Parameters
        -------
        new_selection : bool
        """
        if self._arrow_end_point_selection != new_selection:
            self._arrow_end_point_selection = new_selection
            self.update()

    @Property(bool)
    def arrowStartPoint(self):
        """
        If True, an arrow will be drawn at the start of the first polyline segment.

        Returns
        -------
        bool
        """
        return self._arrow_start_point_selection

    @arrowStartPoint.setter
    def arrowStartPoint(self, new_selection):
        """
        If True, an arrow will be drawn at the start of the first polyline segment.

        Parameters
        -------
        new_selection : bool
        """
        if self._arrow_start_point_selection != new_selection:
            self._arrow_start_point_selection = new_selection
            self.update()

    @Property(bool)
    def arrowMidPoint(self):
        """
        If True, an arrows will be drawn at the midpoints of the segments of the polyline.
        Returns
        -------
        bool
        """
        return self._arrow_mid_point_selection

    @arrowMidPoint.setter
    def arrowMidPoint(self, new_selection):
        """
        If True, an arrows will be drawn at the midpoints of the segments of the polyline.
        Parameters
        -------
        new_selection : bool
        """
        if self._arrow_mid_point_selection != new_selection:
            self._arrow_mid_point_selection = new_selection
            self.update()

    @Property(bool)
    def flipMidPointArrow(self):
        """
        Flips the direction of the midpoint arrows.

        Returns
        -------
        bool
        """
        return self._arrow_mid_point_flipped

    @flipMidPointArrow.setter
    def flipMidPointArrow(self, new_selection):
        """
        Flips the direction of the midpoint arrows.
        
        Parameters
        -------
        new_selection : bool
        """
        if self._arrow_mid_point_flipped != new_selection:
            self._arrow_mid_point_flipped = new_selection
            self.update()

    points = Property("QStringList", getPoints, setPoints, resetPoints)


class PyDMDrawingIrregularPolygon(PyDMDrawingPolyline):
    """
    A widget contains an irregular polygon (arbitrary number of vertices, arbitrary lengths).

    This is a special case of the PyDMDrawingPolyline, adding the requirement that
    the last point is always identical to the first point.

    This widget is created for compatibility with MEDM's *polygon* widget.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the Label
    init_channel : str, optional
        The channel to be used by the widget.
    """

    def getPoints(self):
        return super(PyDMDrawingIrregularPolygon, self).getPoints()

    def resetPoints(self):
        super(PyDMDrawingIrregularPolygon, self).resetPoints()

    def setPoints(self, points):
        verified = self._validator(points)
        if verified is not None:
            if len(verified) > 1:
                if verified[0] != verified[-1]:
                    verified.append(verified[0])  # close the polygon

            if len(verified) < 3:
                logger.error("Must have three or more points")
                return

            self._points = verified
            self.update()

    points = Property("QStringList", getPoints, setPoints, resetPoints)



import locale
import numpy as np
import ast
import shlex
import logging
from functools import partial
from qtpy.QtWidgets import QLineEdit, QMenu, QApplication
from qtpy.QtCore import Property, Q_ENUMS, Qt
from qtpy.QtGui import QFocusEvent
from .. import utilities
from .base import PyDMWritableWidget, TextFormatter, str_types
from .display_format import DisplayFormat, parse_value_for_display

logger = logging.getLogger(__name__)


class PyDMLineEdit(QLineEdit, TextFormatter, PyDMWritableWidget, DisplayFormat):
    Q_ENUMS(DisplayFormat)
    DisplayFormat = DisplayFormat
    """
    A QLineEdit (writable text field) with support for Channels and more
    from PyDM.
    This widget offers an unit conversion menu when users Right Click
    into it.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the Label
    init_channel : str, optional
        The channel to be used by the widget.
    """

    def __init__(self, parent=None, init_channel=None):
        QLineEdit.__init__(self, parent)
        PyDMWritableWidget.__init__(self, init_channel=init_channel)
        self.app = QApplication.instance()
        self._display = None
        self._has_displayed_value_yet = False
        self._scale = 1

        self.returnPressed.connect(self.send_value)
        self.unitMenu = None
        self._display_format_type = self.DisplayFormat.Default
        self._string_encoding = "utf_8"
        self._user_set_read_only = False  # Are we *really* read only?
        if utilities.is_pydm_app():
            self._string_encoding = self.app.get_string_encoding()

    @Property(DisplayFormat)
    def displayFormat(self):
        return self._display_format_type

    @displayFormat.setter
    def displayFormat(self, new_type):
        if self._display_format_type != new_type:
            self._display_format_type = new_type
            # Trigger the update of display format
            self.value_changed(self.value)

    def value_changed(self, new_val):
        """
        Receive and update the PyDMLineEdit for a new channel value

        The actual value of the input is saved as well as the type received.
        This also resets the PyDMLineEdit display text using
        :meth:`.set_display`

        Parameters
        ----------
        value: str, float or int
            The new value of the channel
        """
        super(PyDMLineEdit, self).value_changed(new_val)
        self.set_display()

    def send_value(self):
        """
        Emit a :attr:`send_value_signal` to update channel value.

        The text is cleaned of all units, user-formatting and scale values
        before being sent back to the channel. This function is attached the
        ReturnPressed signal of the PyDMLineEdit
        """
        send_value = str(self.text())
        # Clean text of unit string
        if self._show_units and self._unit and self._unit in send_value:
            send_value = send_value[:-len(self._unit)].strip()
        try:
            if self.channeltype not in [str, np.ndarray, bool]:
                scale = self._scale
                if scale is None or scale == 0:
                    scale = 1.0

                if self._display_format_type in [DisplayFormat.Default, DisplayFormat.String]:
                    if self.channeltype == float:
                        num_value = locale.atof(send_value)
                    else:
                        num_value = self.channeltype(send_value)
                    scale = self.channeltype(scale)
                elif self._display_format_type == DisplayFormat.Hex:
                    num_value = int(send_value, 16)
                elif self._display_format_type == DisplayFormat.Binary:
                    num_value = int(send_value, 2)
                elif self._display_format_type in [DisplayFormat.Exponential, DisplayFormat.Decimal]:
                    num_value = locale.atof(send_value)

                num_value = self.channeltype(num_value / scale)
                self.send_value_signal[self.channeltype].emit(num_value)
            elif self.channeltype == np.ndarray:
                # Arrays will be in the [1.2 3.4 22.214] format
                if self._display_format_type == DisplayFormat.String:
                    self.send_value_signal[str].emit(send_value)
                else:
                    arr_value = list(
                        filter(None, ast.literal_eval(str(shlex.split(send_value.replace("[", "").replace("]", ""))))))
                    arr_value = np.array(arr_value, dtype=self.subtype)
                    self.send_value_signal[np.ndarray].emit(arr_value)
            elif self.channeltype == bool:
                try:
                    val = bool(PyDMLineEdit.strtobool(send_value))
                    self.send_value_signal[bool].emit(val)
                    # might want to add error to application screen
                except ValueError:
                    logger.error("Not a valid boolean: %r", send_value)
            else:
                # Channel Type is String
                # Lets just send what we have after all
                self.send_value_signal[str].emit(send_value)
        except ValueError:
            logger.exception("Error trying to set data '{0}' with type '{1}' and format '{2}' at widget '{3}'."
                             .format(self.text(), self.channeltype, self._display_format_type, self.objectName()))

        self.clearFocus()
        self.set_display()

    def setReadOnly(self, readOnly):
        self._user_set_read_only = readOnly
        super(PyDMLineEdit, self).setReadOnly(True if self._user_set_read_only else not self._write_access)

    def write_access_changed(self, new_write_access):
        """
        Change the PyDMLineEdit to read only if write access is denied
        """
        super(PyDMLineEdit, self).write_access_changed(new_write_access)
        if not self._user_set_read_only:
            super(PyDMLineEdit, self).setReadOnly(not new_write_access)

    def unit_changed(self, new_unit):
        """
        Accept a unit to display with a channel's value

        The unit may or may not be displayed based on the :attr:`showUnits`
        attribute. Receiving a new value for the unit causes the display to
        reset.
        """
        super(PyDMLineEdit, self).unit_changed(new_unit)
        self._scale = 1

    def create_unit_options(self):
        """
        Create the menu for displaying possible unit values

        The menu is filled with possible unit conversions based on the
        current PyDMLineEdit. If either the unit is not found in the by
        the :func:`utilities.find_unit_options` function, or, the
        :attr:`.showUnits` attribute is set to False, the menu will tell
        the user that there are no available conversions
        """
        if self.unitMenu is None:
            self.unitMenu = QMenu('Convert Units', self)
        else:
            self.unitMenu.clear()

        units = utilities.find_unit_options(self._unit)
        if units and self._show_units:
            for choice in units:
                self.unitMenu.addAction(choice,
                                        partial(
                                            self.apply_conversion,
                                            choice
                                            )
                                        )
        else:
            self.unitMenu.addAction('No Unit Conversions found')

    def apply_conversion(self, unit):
        """
        Convert the current unit to a different one

        This function will attempt to find a scalar to convert the current
        unit type to the desired one and reset the display with the new
        conversion.

        Parameters
        ----------
        unit : str
            String name of desired units
        """
        if not self._unit:
            logger.warning("Warning: Attempting to convert PyDMLineEdit unit, but no initial units supplied.")
            return None

        scale = utilities.convert(str(self._unit), unit)
        if scale:
            self._scale = scale * float(self._scale)
            self._unit = unit
            self.update_format_string()
            self.clearFocus()
            self.set_display()
        else:
            logging.warning("Warning: Attempting to convert PyDMLineEdit unit, but '{0}' can not be converted to '{1}'."
                            .format(self._unit, unit))

    def widget_ctx_menu(self):
        """
        Fetch the Widget specific context menu which will be populated with additional tools by `assemble_tools_menu`.

        Returns
        -------
        QMenu or None
            If the return of this method is None a new QMenu will be created by `assemble_tools_menu`.
        """
        self.create_unit_options()

        menu = self.createStandardContextMenu()
        menu.addSeparator()
        menu.addMenu(self.unitMenu)
        return menu

    def set_display(self):
        """
        Set the text display of the PyDMLineEdit.

        The original value given by the PV is converted to a text entry based
        on the current settings for scale value, precision, a user-defined
        format, and the current units. If the user is currently entering a
        value in the PyDMLineEdit the text will not be changed.
        """
        if self.value is None:
            return

        if self.hasFocus():
            return

        new_value = self.value

        if self._display_format_type in [DisplayFormat.Default,
                                         DisplayFormat.Decimal,
                                         DisplayFormat.Exponential,
                                         DisplayFormat.Hex,
                                         DisplayFormat.Binary]:
            if self.channeltype not in (str, np.ndarray):
                try:
                    new_value *= self.channeltype(self._scale)
                except TypeError:
                    logger.error("Cannot convert the value '{0}', for channel '{1}', to type '{2}'. ".format(
                        self._scale, self._channel, self.channeltype))

        new_value = parse_value_for_display(value=new_value,  precision=self.precision,
                                            display_format_type=self._display_format_type,
                                            string_encoding=self._string_encoding,
                                            widget=self)

        self._has_displayed_value_yet = True
        if type(new_value) in str_types:
            self._display = new_value
        else:
            self._display = str(new_value)

        if isinstance(new_value, (int, float)):
            self._display = str(self.format_string.format(new_value))
            self.setText(self._display)
            return

        if self._show_units:
            self._display = "{} {}".format(self._display, self._unit)

        self.setText(self._display)

    def focusInEvent(self, event: QFocusEvent) -> None:
        """
        Checks to see if the line edit has actually received a value before assigning active window or tab focus to it.
        PyQt will automatically give tab focus to the first tab-enabled widget it can on display load. But for this
        widget this behavior can lead to a race condition where if the widget is given focus before the PV has been
        connected long enough to receive a value, then the widget never loads the initial text from the PV.
        """
        if not self._has_displayed_value_yet and (event.reason() == Qt.ActiveWindowFocusReason or event.reason() == Qt.TabFocusReason):
            # Clearing focus ensures that the widget will display the value for the PV
            self.clearFocus()
            return
        super().focusInEvent(event)

    def focusOutEvent(self, event):
        """
        Overwrites the function called when a user leaves a PyDMLineEdit
        without pressing return.  Resets the value of the text field to the
        current channel value.
        """
        if self._display is not None:
            self.setText(self._display)
        super(PyDMLineEdit, self).focusOutEvent(event)

    @staticmethod
    def strtobool(val):
        valid_true = ['Y', 'YES', 'T', 'TRUE', 'ON', '1']
        valid_false = ['N', 'NO', 'F', 'FALSE', 'OFF', '0']

        if val.upper() in valid_true:
            return 1
        elif val.upper() in valid_false:
            return 0
        else:
            raise ValueError("invalid boolean input")



from urllib import parse
import collections
import functools
import logging
import math
import threading

import numpy as np

from qtpy.QtCore import Slot, QThread, Signal, Qt
from qtpy.QtWidgets import QApplication

import pydm
from pydm.data_plugins.plugin import PyDMPlugin, PyDMConnection

logger = logging.getLogger(__name__)


def epics_string(value: np.ndarray, string_encoding: str = "utf-8") -> str:
    """
    Interpret numpy array as a null-terminated string.

    Certain PVs give us char waveforms instead of strings.
    This calculation utility lets us convert these to strings.
    """
    # Stop at the first zero
    # Assume the ndarray is one-dimensional
    value = value.tobytes()
    try:
        value = value[:value.index(0)]
    except (IndexError, ValueError):
        pass
    return value.decode(string_encoding, "replace")  # <-- ignore decoding errors, just in case


def epics_unsigned(value: int, bits: int = 32) -> int:
    """
    Interpret signed integer value as unsigned.

    EPICS channel access is limited to signed types. This helper function
    can be used to succinctly change these to unsigned types for situations
    where the IOC designer is intending for the data type to be unsigned
    but is unable to.
    """
    if value < 0:
        return 2**bits + value
    return value


class CalcThread(QThread):
    eval_env = {'math': math,
                'np': np,
                'numpy': np,
                'epics_string': epics_string,
                'epics_unsigned': epics_unsigned}

    eval_env.update({k: v for k, v in math.__dict__.items() if k[0] != '_'})
    new_data_signal = Signal(dict)
    RESERVED_FIELD = ["update", "expr", "name"]

    def __init__(self, config, *args, **kwargs):
        QThread.__init__(self, *args, **kwargs)
        self.app = QApplication.instance()
        self.app.aboutToQuit.connect(self.requestInterruption)

        self.config = config
        self.listen_for_update = None

        self._calculate = threading.Event()
        self._names = []
        self._channels = []
        self._value = None
        self._values = collections.defaultdict(lambda: None)
        self._connections = collections.defaultdict(lambda: False)
        self._expression = self.config.get('expr', '')[0]

        channels = {}
        for key, channel in self.config.items():
            if key not in CalcThread.RESERVED_FIELD:
                channels[key] = channel[0]

        update = self.config.get('update', None)

        if update is not None:
            self.listen_for_update = update[0].split(',')
            self.listen_for_update = list(map(str.strip, self.listen_for_update))

        for name, channel in channels.items():
            conn_cb = functools.partial(self.callback_conn, name)
            value_cb = functools.partial(self.callback_value, name)
            c = pydm.PyDMChannel(channel, connection_slot=conn_cb,
                                 value_slot=value_cb)
            self._channels.append(c)
            self._names.append(name)

    @property
    def connected(self):
        return all(self._connections.values())

    def _connect(self):
        for ch in self._channels:
            ch.connect()

    def _disconnect(self):
        for ch in self._channels:
            ch.disconnect()

    def _send_update(self, conn, value):
        self.new_data_signal.emit({"connection": conn,
                                   "value": value})

    def run(self):
        self._connect()

        while True:
            self._calculate.wait()
            self._calculate.clear()
            if self.isInterruptionRequested():
                break
            self.calculate_expression()
        self._disconnect()

    def callback_value(self, name, value):
        """
        Callback executed when a channel receives a new value.

        Parameters
        ----------
        name : str
            The channel variable name.
        value : any
            The new value for this channel.

        Returns
        -------
        None
        """
        self._values[name] = value
        if not self.connected:
            logger.debug(
                "Calculation '%s': Not all channels are connected, skipping execution.",
                self.objectName())
            return

        if self.listen_for_update is None or name in self.listen_for_update:
            self._calculate.set()

    def callback_conn(self, name, value):
        """
        Callback executed when a channel connection status is changed.

        Parameters
        ----------
        name : str
            The channel variable name.
        value : bool
            Whether or not this channel is connected.

        """
        self._connections[name] = value
        self._send_update(self.connected, self._value)

    def calculate_expression(self):
        """
        Evaluate the expression defined by the rule and emit the `rule_signal`
        with the new value.
        """
        vals = self._values.copy()
        if any([vals.get(n) is None for n in self._names]):
            logger.debug('Skipping execution as not all values are set.')
            return

        env = dict(CalcThread.eval_env)
        env.update(**vals)
        env.update({'prev_res': self._value})

        try:
            ret = eval(self._expression, env)
            self._value = ret
            self._send_update(self.connected, ret)
        except Exception as e:
            logger.exception("Error while evaluating CalcPlugin connection %s",
                             self.objectName())


class Connection(PyDMConnection):
    def __init__(self, channel, address, protocol=None, parent=None):
        super(Connection, self).__init__(channel, address, protocol, parent)
        self._calc_thread = None
        self.value = None
        self._configuration = {}
        self._waiting_config = True

        self.add_listener(channel)
        self._init_connection()

    def _init_connection(self):
        self.write_access_signal.emit(False)

    def add_listener(self, channel):
        self._setup_calc(channel)
        super(Connection, self).add_listener(channel)
        self.broadcast_value()

    def broadcast_value(self):
        self.connection_state_signal.emit(self.connected)
        if self.value is not None:
            self.new_value_signal[type(self.value)].emit(self.value)

    def _setup_calc(self, channel):
        if not self._waiting_config:
            logger.debug('CalcPlugin connection already configured.')
            return

        try:
            url_data = UrlToPython(channel)
        except ValueError("Not enough information"):
            logger.debug('Invalid configuration for Calc Plugin connection', exc_info=True)
            return

        self._configuration['name'] = url_data.name
        self._configuration.update(url_data.config)
        self._waiting_config = False

        self._calc_thread = CalcThread(self._configuration)
        self._calc_thread.setObjectName("calc_{}".format(url_data.name))
        self._calc_thread.new_data_signal.connect(self.receive_new_data,
                                                  Qt.QueuedConnection)
        self._calc_thread.start()
        return True

    @Slot(dict)
    def receive_new_data(self, data):
        if not data:
            return
        try:
            conn = data.get('connection')
            self.connected = conn
            self.connection_state_signal.emit(conn)
        except KeyError:
            logger.debug('Connection was not available yet for calc.')
        try:
            val = data.get('value')
            self.value = val
            if val is not None:
                self.new_value_signal[type(val)].emit(val)
        except KeyError:
            logger.debug('Value was not available yet for calc.')

    def close(self):
        self._calc_thread.requestInterruption()


class CalculationPlugin(PyDMPlugin):
    protocol = "calc"
    connection_class = Connection

    @staticmethod
    def get_connection_id(channel):
        obj = UrlToPython(channel)
        return obj.name


class UrlToPython:
    def __init__(self, channel):
        self.channel = channel
        self.parsed_address = PyDMPlugin.get_parsed_address(self.channel)
        self.name = None
        self.config = None
        self.get_info()

    def get_info(self):
        """
        Parses a given url into a list and a string.

        Returns
        -------
        A tuple: (<list>, <str>)
        """

        try:
            self.config = parse.parse_qs(self.parsed_address.query.replace("+", "%2B"))
            self.name = self.parsed_address.netloc

            if not self.name or not self.config:
                raise
        except Exception:
            try:
                if not self.name:
                    raise
                logger.debug('Calc Plugin  connection %s got new listener.', self.parsed_address)
                return None, self.name, self.parsed_address
            except Exception:
                msg = "Invalid configuration for Calc Plugin  connection. %s"
                logger.exception(msg, self.parsed_address, exc_info=True)
                raise ValueError("error in Calc Plugin plugin input")

        return True



from qtpy.QtWidgets import QFrame
from qtpy.QtCore import Property
from .base import PyDMWidget


class PyDMFrame(QFrame, PyDMWidget):
    """
    QFrame with support for alarms
    This class inherits from QFrame and PyDMWidget.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the Label
    init_channel : str, optional
        The channel to be used by the widget.
    """
    def __init__(self, parent=None, init_channel=None):
        QFrame.__init__(self, parent)
        PyDMWidget.__init__(self, init_channel=init_channel)

        self._disable_on_disconnect = False
        self.alarmSensitiveBorder = False

    @Property(bool)
    def disableOnDisconnect(self):
        """
        Whether or not the PyDMFrame should be disabled in case the
        channel is disconnected.

        Returns
        -------
        disable : bool
            The configured value
        """
        return self._disable_on_disconnect

    @disableOnDisconnect.setter
    def disableOnDisconnect(self, new_val):
        """
        Whether or not the PyDMFrame should be disabled in case the
        channel is disconnected.

        Parameters
        ----------
        new_val : bool
            The new configuration to use
        """
        if self._disable_on_disconnect != bool(new_val):
            self._disable_on_disconnect = new_val
            self.check_enable_state()

    def check_enable_state(self):
        """
        Checks whether or not the widget should be disable.
        This method also disables the widget and add a Tool Tip
        with the reason why it is disabled.
        """
        if hasattr(self, "_disable_on_disconnect") and self._disable_on_disconnect:
            PyDMWidget.check_enable_state(self)
        else:
            self.setEnabled(True)



import re
import urllib
from .. import config 

def remove_protocol(addr):
    """
    Removes the first occurrence of the protocol string ('://') from the string `addr`

    Parameters
    ----------
    addr : str
        The address from which to remove the address prefix.

    Returns
    -------
    str
    """
    _, addr = protocol_and_address(addr)
    return addr


def protocol_and_address(address):
    """
    Returns the Protocol and Address pieces of a Channel Address

    Parameters
    ----------
    address : str
        The address from which to remove the address prefix.

    Returns
    -------
    protocol : str
        The protocol used. None in case the protocol is not specified.
    addr : str
        The piece of the address without the protocol.
    """
    match = re.match('.*?://', address)
    protocol = None
    addr = address
    if match:
        protocol = match.group(0)[:-3]
        addr = address.replace(match.group(0), '')

    return protocol, addr

def parsed_address(address):
    """
    Returns the given address parsed into a 6-tuple. The parsing is done by urllib.parse.urlparse

    Parameters
    ----------
    address : str
        The address from which to remove the address prefix.

    Returns
    -------
    parsed_address : tuple 
    """
    if type(address) != str:
        return None

    match = re.match('.*?://', address)
    parsed_address = None

    if match:
        parsed_address = urllib.parse.urlparse(address)
    elif config.DEFAULT_PROTOCOL:
        parsed_address = urllib.parse.urlparse(config.DEFAULT_PROTOCOL + '://' + address)

    return parsed_address



import os
import shlex
import subprocess
from functools import partial
import sys
import logging
import warnings
import hashlib
from ast import literal_eval
from qtpy.QtWidgets import QPushButton, QMenu, QMessageBox, QInputDialog, QLineEdit, QWidget
from qtpy.QtGui import QCursor, QIcon, QMouseEvent
from qtpy.QtCore import Property, QSize, Qt, QTimer
from .base import PyDMWidget, only_if_channel_set
from ..utilities import IconFont
from typing import Optional, Union, List

logger = logging.getLogger(__name__)


class PyDMShellCommand(QPushButton, PyDMWidget):
    """
    A QPushButton capable of execute shell commands.

    Parameters
    ----------
    parent : QWidget, optional
        The parent widget for the shell command
    command : str or list, optional
        A string for a single command to run, or a list of strings for multiple commands
    title : str or list, optional
        Title of the command to run, shown in the display. If a list, number of elements must match that of command
    init_channel : str, optional
        The channel to be used by the widget
    """

    DEFAULT_CONFIRM_MESSAGE = "Are you sure you want to proceed?"

    def __init__(self, parent: Optional[QWidget] = None, 
                 command: Optional[Union[str, List[str]]] = None, 
                 title: Optional[Union[str, List[str]]] = None,
                 init_channel: Optional[str] = None) -> None:
        QPushButton.__init__(self, parent)
        PyDMWidget.__init__(self, init_channel=init_channel)
        self.iconFont = IconFont()
        self._icon = self.iconFont.icon("cog")
        self._warning_icon = self.iconFont.icon('exclamation-circle')
        self.setIconSize(QSize(16, 16))
        self.setIcon(self._icon)
        self.setCursor(QCursor(self._icon.pixmap(16, 16)))
        if not title:
            title = []
        if not command:
            command = []
        if isinstance(title, str):
            title = [title]
        if isinstance(command, str):
            command = [command]
        if len(title) > 0 and (len(title) != len(command)):
            raise ValueError("Number of items in 'command' must match number of items in 'title'.")
        self._commands = command
        self._titles = title
        self._menu_needs_rebuild = True
        self._allow_multiple = False
        self.process = None
        self._show_icon = True
        self._redirect_output = False

        self._password_protected = False
        self._password = ""
        self._protected_password = ""
        self.env_var = None

        self._show_confirm_dialog = False
        self._confirm_message = PyDMShellCommand.DEFAULT_CONFIRM_MESSAGE
    
    def confirmDialog(self) -> bool:
        """
        Show the confirmation dialog with the proper message in case
        ```showConfirmMessage``` is True.

        Returns
        -------
        bool
            True if the message was confirmed or if ```showCofirmMessage```
            is False.
        """
        if self._show_confirm_dialog:
            if self._confirm_message == "":
                self._confirm_message = PyDMShellCommand.DEFAULT_CONFIRM_MESSAGE
            
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Question)
            msg.setText(self._confirm_message)
            msg.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
            ret = msg.exec_()

            if ret == QMessageBox.No:
                return False

        return True

    @Property(bool)
    def showConfirmDialog(self) -> bool:
        """
        Wether or not to display a confirmation dialog.

        Returns
        -------
        bool
        """
        return self._show_confirm_dialog

    @showConfirmDialog.setter
    def showConfirmDialog(self, value: bool) -> None:
        """
        Wether or not to display a confirmation dialog.

        Parameters
        ----------
        value : bool
        """
        if self._show_confirm_dialog != value:
            self._show_confirm_dialog = value

    @Property(str)
    def confirmMessage(self) -> str:
        """
        Message to be displayed at the Confirmation dialog.

        Returns
        -------
        str
        """
        return self._confirm_message

    @confirmMessage.setter
    def confirmMessage(self, value: str) -> None:
        """
        Message to be displayed at the Confirmation dialog.

        Parameters
        ----------
        value : str
        """
        if self._confirm_message != value:
            self._confirm_message = value

    @only_if_channel_set
    def check_enable_state(self) -> None:
        """
        override parent method, so this widget does not get disabled when the pv disconnects.
        This method adds a Tool Tip with the reason why it is disabled.
        """
        status = self._connected
        tooltip = self.restore_original_tooltip()
        if not status:
            if tooltip != '':
                tooltip += '\n'
            tooltip += "Alarm PV is disconnected."
            tooltip += '\n'
            tooltip += self.get_address()

        self.setToolTip(tooltip)
        
    @Property(str)
    def environmentVariables(self) -> str:
        """
        Return the environment variables which would be set along with the shell command.

        Returns
        -------
        self.env_var : str
        """
        return self.env_var

    @environmentVariables.setter
    def environmentVariables(self, new_dict: str) -> None:
        """
        Set environment variables which would be set along with the shell command.

        Parameters
        ----------
        new_dict : str
        """
        if self.env_var != new_dict:
            self.env_var = new_dict

    @Property(bool)
    def showIcon(self) -> bool:
        """
        Whether or not we should show the selected Icon.

        Returns
        -------
        bool
        """
        return self._show_icon

    @showIcon.setter
    def showIcon(self, value: bool) -> None:
        """
        Whether or not we should show the selected Icon.

        Parameters
        ----------
        value : bool
        """
        if self._show_icon != value:
            self._show_icon = value

            if self._show_icon:
                self.setIcon(self._icon)
            else:
                self._icon = self.icon()
                self.setIcon(QIcon())

    @Property(bool)
    def redirectCommandOutput(self) -> bool:
        """
        Whether or not we should redirect the output of command to the shell.

        Returns
        -------
        bool
        """
        return self._redirect_output

    @redirectCommandOutput.setter
    def redirectCommandOutput(self, value: bool) -> None:
        """
        Whether or not we should redirect the output of command to the shell.

        Parameters
        ----------
        value : bool

        Returns
        -------
        None.
        """
        if self._redirect_output != value:
            self._redirect_output = value

    @Property(bool)
    def allowMultipleExecutions(self) -> bool:
        """
        Whether or not we should allow the same command
        to be executed even if it is still running.

        Returns
        -------
        bool
        """
        return self._allow_multiple

    @allowMultipleExecutions.setter
    def allowMultipleExecutions(self, value: bool) -> None:
        """
        Whether or not we should allow the same command
        to be executed even if it is still running.

        Parameters
        ----------
        value : bool
        """
        if self._allow_multiple != value:
            self._allow_multiple = value

    @Property('QStringList')
    def titles(self) -> List[str]:
        return self._titles

    @titles.setter
    def titles(self, val: List[str]) -> None:
        self._titles = val
        self._menu_needs_rebuild = True

    @Property('QStringList')
    def commands(self) -> List[str]:
        return self._commands

    @commands.setter
    def commands(self, val: List[str]) -> None:
        if not val:
            self._commands = []
        else:
            self._commands = val
        self._menu_needs_rebuild = True

    @Property(str, designable=False)
    def command(self) -> str:
        """
        DEPRECATED: use the 'commands' property.
        This property simply returns the first command from the 'commands'
        property.
        The shell command to run.

        Returns
        -------
        str
        """
        if len(self.commands) == 0:
            return ""
        return self.commands[0]

    @command.setter
    def command(self, value: str) -> None:
        """
        DEPRECATED: Use the 'commands' property instead.
        This property only has an effect if the 'commands' property is empty.
        If 'commands' is empty, it will be set to a single item list containing
        the value of 'command'.

        Parameters
        ----------
        value : str
        """
        warnings.warn("'PyDMShellCommand.command' is deprecated, "
                      "use 'PyDMShellCommand.commands' instead.")
        if not self._commands:
            if value:
                self.commands = [value]
            else:
                self.commands = []

    @Property(bool)
    def passwordProtected(self) -> bool:
        """
        Whether or not this button is password protected.
        Returns
        -------
        bool
        -------
        """
        return self._password_protected

    @passwordProtected.setter
    def passwordProtected(self, value: bool) -> None:
        """
        Whether or not this button is password protected.
        Parameters
        ----------
        value : bool
        """
        if self._password_protected != value:
            self._password_protected = value

    @Property(str)
    def password(self) -> str:
        """
        Password to be encrypted using SHA256.

        .. warning::
            To avoid issues exposing the password this method
            always returns an empty string.

        Returns
        -------
        str
        """
        return ""

    @password.setter
    def password(self, value: str) -> None:
        """
        Password to be encrypted using SHA256.

        Parameters
        ----------
        value : str
            The password to be encrypted
        """
        if value is not None and value != "":
            sha = hashlib.sha256()
            sha.update(value.encode())
            # Use the setter as it also checks whether the existing password is the same with the
            # new one, and only updates if the new password is different
            self.protectedPassword = sha.hexdigest()

    @Property(str)
    def protectedPassword(self) -> str:
        """
        The encrypted password.

        Returns
        -------
        str
        """
        return self._protected_password

    @protectedPassword.setter
    def protectedPassword(self, value: str) -> None:
        """
        Setter for the encrypted password.

    	Parameters 
    	-------
    	value: str
    	"""
        if self._protected_password != value:
            self._protected_password = value

    def _rebuild_menu(self) -> None:
        if not any(self._commands):
            self._commands = []
        if not any(self._titles):
            self._titles = []
        if len(self._commands) == 0:
            self.setEnabled(False)
        if len(self._commands) <= 1:
            self.setMenu(None)
            self._menu_needs_rebuild = False
            return
        menu = QMenu(self)
        for i, command in enumerate(self._commands):
            if i >= len(self._titles):
                title = command
            else:
                title = self._titles[i]
            action = menu.addAction(title)
            action.triggered.connect(partial(self.execute_command, command))
        self.setMenu(menu)
        self._menu_needs_rebuild = False

    def mousePressEvent(self, event: QMouseEvent) -> None:
        if self._menu_needs_rebuild:
            self._rebuild_menu()
        super(PyDMShellCommand, self).mousePressEvent(event)

    def mouseReleaseEvent(self, mouse_event: QMouseEvent) -> None:
        """
        mouseReleaseEvent is called when a mouse button is released.
        This means that if the user presses the mouse inside your widget,
        then drags the mouse somewhere else before releasing the mouse
        button, your widget receives the release event.

        Parameters
        ----------
        mouse_event :
        """
        if mouse_event.button() != Qt.LeftButton:
            return super(PyDMShellCommand, self).mouseReleaseEvent(mouse_event)
        if self.menu() is not None:
            return super(PyDMShellCommand, self).mouseReleaseEvent(mouse_event)
        assert len(self.commands) == 1, "More than one command present, but no menu created."
        self.execute_command(self.commands[0])
        super(PyDMShellCommand, self).mouseReleaseEvent(mouse_event)

    def show_warning_icon(self) -> None:
        """ Show the warning icon.  This is called when a shell command fails
        (i.e. exits with nonzero status) """
        self.setIcon(self._warning_icon)
        QTimer.singleShot(5000, self.hide_warning_icon)

    def hide_warning_icon(self) -> None:
        """ Hide the warning icon.  This is called on a timer after the warning
        icon is shown."""
        if self._show_icon:
            self.setIcon(self._icon)
        else:
            self.setIcon(QIcon())

    def validate_password(self) -> bool:
        """
        If the widget is ```passwordProtected```, this method will propmt
        the user for the correct password.

        Returns
        -------
        bool
            True in case the password was correct of if the widget is not
            password protected.
        """
        if not self._password_protected:
            return True

        pwd, ok = QInputDialog().getText(None, "Authentication", "Please enter your password:",
                                         QLineEdit.Password, "")
        pwd = str(pwd)
        if not ok or pwd == "":
            return False

        sha = hashlib.sha256()
        sha.update(pwd.encode())
        pwd_encrypted = sha.hexdigest()
        if pwd_encrypted != self._protected_password:
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Critical)
            msg.setText("Invalid password.")
            msg.setWindowTitle("Error")
            msg.setStandardButtons(QMessageBox.Ok)
            msg.setDefaultButton(QMessageBox.Ok)
            msg.setEscapeButton(QMessageBox.Ok)
            msg.exec_()
            return False
        return True

    def execute_command(self, command: str) -> None:
        """
        Execute the shell command given by ```command```.
        The process is available through the ```process``` member.

        Parameters
        ----------
        command : str 
            Shell command
        """
        if not command:
            logger.info("The command is not set, so no command was executed.")
            return

        if not self.validate_password():
            return None

        if not self.confirmDialog():
            return None

        if (self.process is None or self.process.poll() is not None) or self._allow_multiple:
            cmd = os.path.expanduser(os.path.expandvars(command))
            args = shlex.split(cmd, posix='win' not in sys.platform)
            try:
                logger.debug("Launching process: %s", repr(args))
                stdout = subprocess.PIPE

                if self.env_var:
                    env_var = literal_eval(self.env_var)
                else:
                    env_var = None

                if self._redirect_output:
                    stdout = None
                self.process = subprocess.Popen(
                    args, stdout=stdout, stderr=subprocess.PIPE, env=env_var)
            except Exception as exc:
                self.show_warning_icon()
                logger.error("Error in shell command: %s", exc)
        else:
            logger.error("Command '%s' already active.", command)




import logging
import os
import platform
import tempfile

from qtpy import QtWidgets

from ...utilities import (find_display_in_path, is_pydm_app, is_qt_designer,
                          log_failures, path_info, which)

logger = logging.getLogger(__name__)


def test_is_pydm_app(qapp):
    assert is_pydm_app(qapp)


def test_negative_is_pydm_app():
    assert not is_pydm_app(QtWidgets.QLabel())


def test_is_qt_designer():
    assert not is_qt_designer()


def test_path_info():
    dir_name, file_name, args = path_info('/bin/ls -l -a')
    assert (dir_name == '/bin')
    assert (file_name == 'ls')
    assert (args == ['-l', '-a'])

    dir_name, file_name, args = path_info('/bin/ls')
    assert (dir_name == '/bin')
    assert (file_name == 'ls')
    assert (args == [])


def test_find_display_in_path():
    temp, file_path = tempfile.mkstemp(suffix=".ui", prefix="display_")
    direc, fname, _ = path_info(file_path)
    # Try to find the file as is... is should not find it.
    assert(find_display_in_path(fname) is None)

    # Try to find the file passing the path
    disp_path = find_display_in_path(fname, mode=None, path=direc)
    assert(disp_path == file_path)

    # Try to find the file passing the path but relative name
    rel_name = ".{}{}".format(os.sep, fname)
    expected = "{}{}{}".format(direc, os.sep, rel_name)
    disp_path = find_display_in_path(rel_name, mode=None, path=direc)
    assert (disp_path == expected)


def test_which():
    if platform.system() == 'Windows':
        out = which('ping')
        assert out.lower() == r'c:\windows\system32\ping.exe'

        out = which(r'C:\Windows\System32\PING.EXE')
        assert out.lower() == r'c:\windows\system32\ping.exe'
    else:
        out = which('ls')
        assert 'ls' in out

        out = which('/bin/ls')
        assert 'ls' in out

    out = which('non_existant_binary')
    assert out is None

    out = which('/bin/non_existant_binary')
    assert out is None

    out = which('ls', path='')
    assert out is None


def test_log_failure_capture(caplog):
    @log_failures(logger)
    def inner():
        raise ValueError()

    with caplog.at_level(logging.WARNING):
        result = inner()

    assert result is None

    assert len(caplog.records) == 1
    assert "Failed to run inner" in str(caplog.records[0])


def test_log_failure_pass(caplog):
    @log_failures(logger)
    def inner():
        return 5

    with caplog.at_level(logging.WARNING):
        result = inner()

    assert result == 5
    assert not caplog.records



import os
import tempfile
import pytest
import json

from ...utilities.macro import substitute_in_file, parse_macro_string


@pytest.mark.parametrize("text, macros, expected", [
    (
            'This is a ${what} to ensure that macros work. ${A} ${B} ${C}',
            {'what': 'test', 'A': 'X', 'B': 'Y', 'C': 'Z'},
            'This is a test to ensure that macros work. X Y Z'
    ),
    (
            'This is a ${what} to ensure that macros work. ${A} ${B} ${C}',
            {'what': 'test', 'A': 'X', 'B': 'Y', 'C': 'Z', 'D': 'W'},
            'This is a test to ensure that macros work. X Y Z'
    ),
    (
            'This is a ${what} to ensure that macros work. ${A} ${B} ${C}',
            {'what': 'test', 'A': 'X', 'B': 'Y'},
            'This is a test to ensure that macros work. X Y ${C}'
    )
])
def test_substitute_in_file(text, macros, expected):
    fd, name = tempfile.mkstemp(text=True)
    # use a context manager to open the file at that path and close it again
    with open(name, 'w') as f:
        f.write(text)
    # close the file descriptor
    os.close(fd)

    nf = substitute_in_file(name, macros)
    nt = nf.read()
    assert (nt == expected)

@pytest.mark.parametrize("macro_string, expected_dict", [
    ('{"A": "1", "B": "2"}', {"A": "1", "B": "2"}),
    ("A=1,B=2", {"A": "1", "B": "2"}),
    ("A=$(other_macro),B=2,C=3", {"A": "$(other_macro)", "B": "2", "C": "3"}),
    ("A=$(other_macro=3)", {"A": "$(other_macro=3)"}),
    ("TITLE='1,2', B=2, C=3", {"TITLE": "1,2", "B": "2", "C": "3"}),
    ("TITLE=1\,2,B=2,C=3", {"TITLE": "1,2", "B": "2", "C": "3"}),
    ('TITLE="e=mc^2",B=2,C=3', {"TITLE": "e=mc^2", "B": "2", "C": "3"}),
    ('', {}),
    (None, {})
])
def test_macro_parser(macro_string, expected_dict):
    """
    Test the parser, using a couple of normal cases, and a bunch of perverse
    macro strings that only some insane macro genius (or huge macro idiot) 
    would ever attempt.
    """
    assert parse_macro_string(macro_string) == expected_dict



"""
Plugin to handle EPICS connections using pyca through psp.Pv.
This is used instead of pyepics for better performance.
"""
import numpy as np
import pyca
from psp.Pv import Pv
from qtpy.QtCore import Slot, Qt, QTimer
from pydm import data_plugins
from pydm.data_plugins.plugin import PyDMPlugin, PyDMConnection

# Map how we will interpret EPICS types in python.
type_map = dict(
    DBF_STRING=str,
    DBF_CHAR=str,
    DBF_UCHAR=str,
    DBF_SHORT=int,
    DBF_USHORT=int,
    DBF_LONG=int,
    DBF_ULONG=int,
    DBF_FLOAT=float,
    DBF_DOUBLE=float,
    DBF_ENUM=int,
    DBF_MENU=None,
    DBF_DEVICE=None,
    DBF_INLINK=None,
    DBF_OUTLINK=None,
    DBF_FWDLINK=None,
    DBF_NOACCESS=None,
)

# .SCAN mapping to override throttle
scan_list = [
    float("inf"),  # passive
    float("inf"),  # event
    float("inf"),  # IO Intr
    10.0,
    5.0,
    2.0,
    1.0,
    0.5,
    0.2,
    0.1,
]


def generic_con_cb(pv_obj):
    """
    Create a generic callback to set up a monitor on connection.

    :param pv_obj: Object representing the EPICS PV data source. This object
                   needs to have not been connected yet.
    :type pv_obj:  Pv
    :rtype: function(is_connected)
    """

    def cb(is_connected):
        if is_connected and not pv_obj.ismonitored:
            pv_obj.monitor()

    return cb


def generic_mon_cb(source, signal):
    """
    Create a generic callback for sending a signal from source value.

    :param source: Object representing the EPICS PV data source. This object
                   needs to be properly initialized and monitored.
    :type source:  Pv
    :param signal: Signal to send the value out on. Check the base
                   :class:`PyDMConnection` class for available signals.
    :type signal:  Signal
    :rtype: function(errors=None)
    """

    def cb(e=None):
        if e is None:
            try:
                signal.emit(source.value)
            except AttributeError:
                raise

    return cb


def setup_pv(pvname, con_cb=None, mon_cb=None, rwaccess_cb=None, signal=None, mon_cb_once=False, control=False):
    """
    Initialize an EPICS PV using psp with proper callbacks.

    :param pvname: EPICS PV name
    :type pvname:  str
    :param con_cb: Connection callback. If left as None and provided with
                    signal, emit our value from signal as the callback.
    :type con_cb:  function(isconnected=None)
    :param mon_cb: Monitor callback. If left as None and provided with signal,
                   emit our value from signal as the callback.
    :type mon_cb:  function(errors=None)
    :param rwaccess_cb: Read/write access state callback.
    :type rwaccess_cb: function(access_state)
    :param signal: Signal to emit our value on as the default callback when
                   con_cb or mon_cb are left as None. Check the base
                   :class:`PyDMConnection` class for available signals.
    :type signal:  Signal
    :param mon_cb_once: True if we only want the monitor callback to run once.
    :type mon_cb_once: bool
    :param control: True if we want to monitor control values
    :type control: bool
    :rtype: Pv
    """
    pv = Pv(pvname, use_numpy=True, control=control)

    if signal is None:
        default_mon_cb = lambda e: None
    else:
        default_mon_cb = generic_mon_cb(pv, signal)

    pv.add_connection_callback(con_cb or generic_con_cb(pv))
    pv.add_monitor_callback(mon_cb or default_mon_cb, once=mon_cb_once)
    if rwaccess_cb:
        pv.add_rwaccess_callback(rwaccess_cb)
    pv.connect(None)
    return pv


class Connection(PyDMConnection):
    """
    Class that manages channel access connections using pyca through psp.
    See :class:`PyDMConnection` class.
    """

    def __init__(self, channel, pv, protocol=None, parent=None):
        """
        Instantiate Pv object and set up the channel access connections.

        :param channel: :class:`PyDMChannel` object as the first listener.
        :type channel:  :class:`PyDMChannel`
        :param pv: Name of the pv to connect to.
        :type pv:  str
        :param parent: PyQt widget that this widget is inside of.
        :type parent:  QWidget
        """
        super(Connection, self).__init__(channel, pv, protocol, parent)
        self.python_type = None
        self.pv = setup_pv(pv,
                           con_cb=self.connected_cb,
                           mon_cb=self.monitor_cb,
                           rwaccess_cb=self.rwaccess_cb,
                           control=True)
        self.enums = None
        self.sevr = None
        self.ctrl_llim = None
        self.ctrl_hlim = None
        self.alarm_hlim = None
        self.alarm_llim = None
        self.warn_hlim = None
        self.warn_llim = None

        self.units = None
        self.prec = None
        self.time = None
        self.count = None
        self.epics_type = None
        self.read_access = False
        self.write_access = False
        # Auxilliary info to help with throttling
        self.scan_pv = setup_pv(pv + ".SCAN", mon_cb=self.scan_pv_cb,
                                mon_cb_once=True)
        self.throttle = QTimer(self)
        self.throttle.timeout.connect(self.throttle_cb)

        self.add_listener(channel)

    def connected_cb(self, isconnected):
        """
        Callback to run whenever the connection state of our pv changes.

        :param isconnected: True if we are connected, False otherwise.
        :type isconnected:  bool
        """
        self.connected = isconnected
        self.send_connection_state(isconnected)
        if isconnected:
            self.epics_type = self.pv.type()
            self.count = self.pv.count or 1

            # Get the control info for the PV.
            self.pv.get_data(True, -1.0, self.count)
            pyca.flush_io()
            if self.epics_type == "DBF_ENUM":
                self.pv.get_enum_strings(-1.0)
            if not self.pv.ismonitored:
                self.pv.monitor()
            self.python_type = type_map.get(self.epics_type)
            if self.python_type is None:
                raise Exception("Unsupported EPICS type {0} for pv {1}".format(
                    self.epics_type, self.pv.name))

    def monitor_cb(self, e=None):
        """
        Callback to run whenever the value of our pv changes.

        :param e: Error state. Should be None under normal circumstances.
        """
        if e is None:
            self.send_new_value(self.pv.value)

    def rwaccess_cb(self, read_access, write_access):
        """
        Callback to run when the access state of our pv changes.

        :param read_access: Whether or not the PV is readable.
        :param write_access: Whether or not the PV is writeable.
        """
        self.read_access = read_access
        self.write_access = write_access
        self.send_access_state()

    def throttle_cb(self):
        """
        Callback to run when the throttle timer times out.
        """
        self.send_new_value(self.pv.get())

    def timestamp(self):
        try:
            secs, nanos = self.pv.timestamp()
        except KeyError:
            return None
        return float(secs + nanos / 1.0e9)

    def send_new_value(self, value=None):
        """
        Send a value to every channel listening for our Pv.

        :param value: Value to emit to our listeners.
        :type value:  int, float, str, or np.ndarray, depending on our record
                      type.
        """
        if self.python_type is None:
            return

        if self.enums is None:
            try:
                self.update_enums()
            except KeyError:
                self.pv.get_enum_strings(-1.0)

        if self.pv.severity is not None and self.pv.severity != self.sevr:
            self.sevr = self.pv.severity
            self.new_severity_signal.emit(self.sevr)

        try:
            prec = self.pv.data['precision']
        except KeyError:
            pass
        else:
            if self.prec != prec:
                self.prec = prec
                self.prec_signal.emit(int(self.prec))

        try:
            units = self.pv.data['units']
        except KeyError:
            pass
        else:
            if self.units != units:
                self.units = units
                self.unit_signal.emit(self.units.decode(encoding='ascii') if isinstance(self.units, bytes) else self.units)

        time = self.timestamp()

        if time is not None and self.time != time:
            self.time = time
            self.timestamp_signal.emit(self.time)

        try:
            ctrl_llim = self.pv.data['ctrl_llim']
        except KeyError:
            pass
        else:
            if self.ctrl_llim != ctrl_llim:
                self.ctrl_llim = ctrl_llim
                self.lower_ctrl_limit_signal.emit(self.ctrl_llim)

        try:
            ctrl_hlim = self.pv.data['ctrl_hlim']
        except KeyError:
            pass
        else:
            if self.ctrl_hlim != ctrl_hlim:
                self.ctrl_hlim = ctrl_hlim
                self.upper_ctrl_limit_signal.emit(self.ctrl_hlim)

        try:
            alarm_hlim = self.pv.data['alarm_hlim']
        except KeyError:
            pass
        else:
            if self.alarm_hlim != alarm_hlim:
                self.alarm_hlim = alarm_hlim
                self.upper_alarm_limit_signal.emit(self.alarm_hlim)

        try:
            alarm_llim = self.pv.data['alarm_llim']
        except KeyError:
            pass
        else:
            if self.alarm_llim != alarm_llim:
                self.alarm_llim = alarm_llim
                self.lower_alarm_limit_signal.emit(self.alarm_llim)

        try:
            warn_hlim = self.pv.data['warn_hlim']
        except KeyError:
            pass
        else:
            if self.warn_hlim != warn_hlim:
                self.warn_hlim = warn_hlim
                self.upper_warning_limit_signal.emit(self.warn_hlim)

        try:
            warn_llim = self.pv.data['warn_llim']
        except KeyError:
            pass
        else:
            if self.warn_llim != warn_llim:
                self.warn_llim = warn_llim
                self.lower_warning_limit_signal.emit(self.warn_llim)

        if self.count > 1:
            self.new_value_signal[np.ndarray].emit(value)
        else:
            self.new_value_signal[self.python_type].emit(self.python_type(value))

    def send_ctrl_vars(self):
        if self.enums is None:
            try:
                self.update_enums()
            except KeyError:
                self.pv.get_enum_strings(-1.0)
        else:
            self.enum_strings_signal.emit(self.enums)

        if self.pv.severity != self.sevr:
            self.sevr = self.pv.severity
        self.new_severity_signal.emit(self.sevr)

        if self.prec is None:
            try:
                self.prec = self.pv.data['precision']
            except KeyError:
                pass
        if self.prec is not None:
            self.prec_signal.emit(int(self.prec))

        if self.time is None:
            self.time = self.timestamp()

        if self.time is not None:
            self.timestamp_signal.emit(self.time)

        if self.units is None:
            try:
                self.units = self.pv.data['units']
            except KeyError:
                pass
        if self.units:
            if isinstance(self.units, bytes):
                self.units = self.units.decode(encoding='ascii')
            self.unit_signal.emit(self.units)

        if self.ctrl_llim is None:
            try:
                self.ctrl_llim = self.pv.data['ctrl_llim']
            except KeyError:
                pass
        if self.ctrl_llim is not None:
            self.lower_ctrl_limit_signal.emit(self.ctrl_llim)

        if self.ctrl_hlim is None:
            try:
                self.ctrl_hlim = self.pv.data['ctrl_hlim']
            except KeyError:
                pass
        if self.ctrl_hlim is not None:
            self.upper_ctrl_limit_signal.emit(self.ctrl_hlim)

        if self.alarm_hlim is None:
            try:
                self.alarm_hlim = self.pv.data['alarm_hlim']
            except KeyError:
                pass
        if self.alarm_hlim is not None:
            self.upper_alarm_limit_signal.emit(self.alarm_hlim)

        if self.alarm_llim is None:
            try:
                self.alarm_llim = self.pv.data['alarm_llim']
            except KeyError:
                pass
        if self.alarm_llim is not None:
            self.lower_alarm_limit_signal.emit(self.alarm_llim)

        if self.warn_hlim is None:
            try:
                self.warn_hlim = self.pv.data['warn_hlim']
            except KeyError:
                pass
        if self.warn_hlim is not None:
            self.upper_warning_limit_signal.emit(self.warn_hlim)

        if self.warn_llim is None:
            try:
                self.warn_llim = self.pv.data['warn_llim']
            except KeyError:
                pass
        if self.warn_llim is not None:
            self.lower_warning_limit_signal.emit(self.warn_llim)

    def send_connection_state(self, conn=None):
        """
        Send an update on our connection state to every listener.

        :param conn: True if we are connected, False if we are disconnected.
        :type conn:  bool
        """
        self.connection_state_signal.emit(conn)

    def send_access_state(self):
        if data_plugins.is_read_only():
            self.write_access_signal.emit(False)
            return
        self.write_access_signal.emit(self.write_access)

    def update_enums(self):
        """
        Send an update on our enum strings to every listener, if this is an
        enum record.
        """
        if self.epics_type == "DBF_ENUM":
            if self.enums is None:
                self.enums = tuple(b.decode(encoding='ascii') if isinstance(b, bytes) else b for b in self.pv.data["enum_set"])
            self.enum_strings_signal.emit(self.enums)

    @Slot(int)
    @Slot(float)
    @Slot(str)
    @Slot(np.ndarray)
    def put_value(self, value):
        """
        Set our PV's value in EPICS.

        :param value: The value we'd like to put to our PV.
        :type value:  int or float or str or np.ndarray, depending on our
                      record type.
        """
        if self.count == 1:
            value = self.python_type(value)
        try:
            self.pv.put(value)
        except pyca.caexc as e:
            print("pyca error: {}".format(e))

    @Slot(np.ndarray)
    def put_waveform(self, value):
        """
        Set a PV's waveform value in EPICS. This is a deprecated function kept
        temporarily for compatibility with old code.

        :param value: The waveform value we'd like to put to our PV.
        :type value:  np.ndarray
        """
        self.put_value(value)

    def scan_pv_cb(self, e=None):
        """
        Call set_throttle once we have a value from the scan_pv. We need this
        value inside set_throttle to decide if we can ignore the throttle
        request (i.e. our pv updates more slowly than our throttle)

        :param e: Error state. Should be None under normal circumstances.
        """
        if e is None:
            self.pv.wait_ready()
            count = self.pv.count or 1
            if count > 1:
                max_data_rate = 1000000.  # bytes/s
                bytes = self.pv.value.itemsize  # bytes
                throttle = max_data_rate / (bytes * count)  # Hz
                if throttle < 120:
                    self.set_throttle(throttle)

    @Slot(int)
    @Slot(float)
    def set_throttle(self, refresh_rate):
        """
        Throttle our update rate. This is useful when the data is large (e.g.
        image waveforms). Set to zero to disable throttling.

        :param delay: frequency of pv updates
        :type delay:  float or int
        """
        try:
            scan = scan_list[self.scan_pv.value]
        except Exception:
            scan = float("inf")
        if 0 < refresh_rate < 1 / scan:
            self.pv.monitor_stop()
            self.throttle.setInterval(1000.0 / refresh_rate)
            self.throttle.start()
        else:
            self.throttle.stop()
            if not self.pv.ismonitored:
                self.pv.monitor()

    def add_listener(self, channel):
        """
        Connect a channel's signals and slots with this object's signals and slots.

        :param channel: The channel to connect.
        :type channel:  :class:`PyDMChannel`
        """
        super(Connection, self).add_listener(channel)
        # If we are adding a listener to an already existing PV, we need to
        # manually send the signals indicating that the PV is connected, what
        # the latest value is, etc.
        if self.pv.isconnected and self.pv.isinitialized:
            self.send_connection_state(conn=True)
            self.monitor_cb()
            try:
                self.update_enums()
            except KeyError:
                self.pv.get_enum_strings(-1.0)
            self.send_access_state()
            self.send_ctrl_vars()
        if channel.value_signal is not None:
            try:
                channel.value_signal[str].connect(self.put_value, Qt.QueuedConnection)
            except KeyError:
                pass
            try:
                channel.value_signal[int].connect(self.put_value, Qt.QueuedConnection)
            except KeyError:
                pass
            try:
                channel.value_signal[float].connect(self.put_value, Qt.QueuedConnection)
            except KeyError:
                pass
            try:
                channel.value_signal[np.ndarray].connect(self.put_value, Qt.QueuedConnection)
            except KeyError:
                pass

    def close(self):
        """
        Clean up.
        """
        self.throttle.stop()
        self.pv.monitor_stop()
        self.pv.disconnect()
        self.scan_pv.monitor_stop()
        self.scan_pv.disconnect()


class PSPPlugin(PyDMPlugin):
    """
    Class to define our protocol and point to our :class:`Connection` Class
    """
    # NOTE: protocol is intentionally "None" to keep this plugin from getting directly imported.
    # If this plugin is chosen as the One True EPICS Plugin in epics_plugin.py, the protocol will
    # be properly set before it is used.
    protocol = None
    connection_class = Connection



from pathlib import Path
from qtpy.QtWidgets import QTextBrowser, QVBoxLayout, QWidget
from qtpy.QtCore import Qt
from typing import Optional


class HelpWindow(QWidget):
    """
    A window for displaying a help file for a PyDM display

    Parameters
    ----------
    help_file_path : str
        The path to the help file to be displayed
    """
    def __init__(self, help_file_path: str, parent: Optional[QWidget] = None):
        super().__init__(parent, Qt.Window)
        self.resize(500, 400)

        path = Path(help_file_path)
        self.setWindowTitle(f'Help for {path.stem}')

        self.display_content = QTextBrowser()

        with open(help_file_path) as file:
            if path.suffix == '.txt':
                self.display_content.setText(file.read())
            else:
                self.display_content.setHtml(file.read())

        self.vBoxLayout = QVBoxLayout()
        self.vBoxLayout.addWidget(self.display_content)
        self.setLayout(self.vBoxLayout)



# Unit Tests for the PyDMLineEdit Widget

import pytest

import functools
import numpy as np
from logging import ERROR

from qtpy.QtCore import QEvent, Qt
from qtpy.QtGui import QFocusEvent
from qtpy.QtWidgets import QMenu
from ...widgets.line_edit import PyDMLineEdit
from ...data_plugins import set_read_only, is_read_only
from ...utilities import is_pydm_app, find_unit_options
from ...widgets.display_format import DisplayFormat, parse_value_for_display


# --------------------
# POSITIVE TEST CASES
# --------------------

@pytest.mark.parametrize("init_channel", [
    "CA://MTEST",
    "",
    None,
])
def test_construct(qtbot, init_channel):
    """
    Test the widget construct.
    Expectations:
    All parameters have the correct default value.

    Parameters
    ----------
     qtbot : fixture
        Window for widget testing
    init_channel : str
        The data channel to be used by the widget

    """
    pydm_lineedit = PyDMLineEdit(init_channel=init_channel)
    qtbot.addWidget(pydm_lineedit)

    if init_channel:
        assert pydm_lineedit.channel == str(init_channel)
    else:
        assert pydm_lineedit.channel is None
    assert pydm_lineedit._display is None
    assert pydm_lineedit._scale == 1
    assert pydm_lineedit._prec == 0
    assert pydm_lineedit.showUnits == False

    assert pydm_lineedit.unitMenu is None
    pydm_lineedit.widget_ctx_menu()

    assert isinstance(pydm_lineedit.unitMenu, QMenu) and pydm_lineedit.unitMenu.title() == "Convert Units"
    assert pydm_lineedit.displayFormat == pydm_lineedit.DisplayFormat.Default
    assert (pydm_lineedit._string_encoding == pydm_lineedit.app.get_string_encoding()
            if is_pydm_app() else "utf_8")

    assert find_action_from_menu(pydm_lineedit.unitMenu, "No Unit Conversions found")


@pytest.mark.parametrize("display_format", [
    (DisplayFormat.Default),
    (DisplayFormat.Exponential),
    (DisplayFormat.String),
    (DisplayFormat.Binary),
    (DisplayFormat.Decimal),
    (DisplayFormat.Hex),
])
def test_change_display_format_type(qtbot, display_format):
    """
    Test the widget's DisplayFormat property and setter.

    Expectations:
    The widget can set the DisplayFormat property, and retrieve it correctly.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    display_format : DisplayFormat
        The new DisplayFormat value to set
    """
    pydm_lineedit = PyDMLineEdit()
    qtbot.addWidget(pydm_lineedit)

    pydm_lineedit.displayFormat = display_format
    assert pydm_lineedit.displayFormat == display_format


@pytest.mark.parametrize("value, display_format, precision, scale, unit, show_unit, expected_display", [
    (123, DisplayFormat.Default, 3, 1, "s", True, "123.000 s"),
    (123.47, DisplayFormat.Decimal, 3, 2, "seconds", False, "246.940"),
    (123.4567, DisplayFormat.String, 2, 1, "", False, "123.46"),
    (123.4567, DisplayFormat.Decimal, 1, 1, "", False, "123.5"),
    (1e2, DisplayFormat.Exponential, 2, 2, "light years", True, "2.00e+02 light years"),
    (0x1FF, DisplayFormat.Hex, 0, 1, "Me", True, "0x1ff Me"),
    (0b100, DisplayFormat.Binary, 0, 1, "KB", True, "0b100 KB"),
    (np.array([123, 456]), DisplayFormat.Default, 3, 2, "light years", True, "[123 456] light years"),
])
def test_value_change(qtbot, signals, value, display_format, precision, scale, unit, show_unit, expected_display):
    """
    Test changing the value to be displayed by the widget, given the value's display format, precision, scale, and unit.

    Expectations:
    The widget displays the new value properly, taken into account the value's display format, precision, scale, and
    unit.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    signals : fixture
        The signals fixture, which provides access signals to be bound to the appropriate slots
    value : int, float, bin, hex, numpy.array
        The value to be displayed by the widget
    display_format : DisplayFormat
        The value's display format (Default, Decimal, Exponential, Hex, Binary)
    precision : int
        The number of decimal places in the displayed value
    scale : int
        The factor to reduce or magnify the given value
    unit : str
        The value unit to be displayed
    show_unit : bool
        True if the value unit is to be displayed. False otherwise
    expected_display : str
        The expected display contents to be compared to the actual widget's displayed content
    """
    pydm_lineedit = PyDMLineEdit()
    qtbot.addWidget(pydm_lineedit)

    pydm_lineedit.displayFormat = display_format
    pydm_lineedit._prec = precision
    pydm_lineedit._scale = scale
    pydm_lineedit.channeltype = type(value)
    pydm_lineedit._unit = unit
    pydm_lineedit.showUnits = show_unit

    signals.new_value_signal[type(value)].connect(pydm_lineedit.channelValueChanged)
    signals.new_value_signal[type(value)].emit(value)

    assert pydm_lineedit._display == expected_display


@pytest.mark.parametrize("init_value, user_typed_value, display_format, precision, scale, unit,"
                         "show_units, expected_received_value, expected_display_value", [
    ("abc", "cdf", DisplayFormat.Default, 3, 5, "s", True, "abc", "cdf s"),
    ("abc", "cdf", DisplayFormat.String, 3, 5, "s", False, "abc", "cdf"),
    (np.array([65, 66]), "[C D]", DisplayFormat.Default, 0, 10, "light years", True, "[C D]",
    "[C D] light years"),
    (np.array(["A", "B"]), "[C D]", DisplayFormat.String, 0, 10, "ms", True, "[C D]", "[C D] ms"),
    (np.array(["A", "B"]), np.array(["C", "D"]), DisplayFormat.String, 0, 10, "ms", True, "C   D    ", "C   D    ms"),
    (np.array(["A", "B"]), np.array(["C", "D"]), DisplayFormat.Default, 0, 10, "ms", True, "['C' 'D']", "['C' 'D'] ms"),
    (0, 10, DisplayFormat.Default, 0, 1.0, "V", True, 10, "10 V"),
])
def test_send_value(qtbot, signals, init_value, user_typed_value, display_format, precision, scale, unit,
                    show_units, expected_received_value, expected_display_value):
    """
    Test sending the value to the channel, and the displayed value by the widget.

    Expectations:

    1. The value sent to the channel will be affected by the precision only
    2. The value displayed by the widget will be affected by the scale, and show unit Boolean flag

    Parameters
    ----------
    qtbot : fixture
        pytest-qt window for widget testing
    signals : fixture
        The signals fixture, which provides access signals to be bound to the appropriate slots
    init_value : int, float, hex, bin, str, numpy.array
        The initial value currently assigned to the widget
    user_typed_value : str
        The new value as provided (typed in) by the user
    display_format : DisplayFormat
        The display format of the current value, only str allowed
    precision : int
        The number of decimal places to consider for the new value
    scale : int
        The factor to reduce or magnify the new value
    unit : str
        The unit of the new value
    show_units : bool
        True if the value unit is to be displayed. False otherwise
    expected_received_value : int, float, hex, bin, str, numpy.array
        The expected new value to send to the channel
    expected_display_value : str
        The expected new content to be displayed by the widget
    """
    pydm_lineedit = PyDMLineEdit()
    qtbot.addWidget(pydm_lineedit)

    pydm_lineedit.value = init_value
    pydm_lineedit.channeltype = type(init_value)
    pydm_lineedit.displayFormat = display_format
    pydm_lineedit._prec = precision
    pydm_lineedit._scale = scale
    pydm_lineedit._unit = unit
    pydm_lineedit.showUnits = show_units

    signals.new_value_signal[type(user_typed_value)].connect(pydm_lineedit.channelValueChanged)
    signals.new_value_signal[type(user_typed_value)].emit(user_typed_value)

    # Besides receiving the new channel value, simulate the update of the new value to the widget by connecting the
    # channelValueChanged slot to the same signal
    signal_type = type(user_typed_value)
    if signal_type == np.ndarray:
        signal_type = str
    pydm_lineedit.send_value_signal[signal_type].connect(signals.receiveValue)
    pydm_lineedit.send_value_signal[signal_type].connect(pydm_lineedit.channelValueChanged)
    pydm_lineedit.send_value()

    # Python 2.7 produces the strings without the spaces, but Python 3.x does not. So, remove all the spaces, and
    # compare just the characters (in the right order)
    assert pydm_lineedit.displayText().replace(" ", "") == expected_display_value.replace(" ", "")

    if all(x in (int, float) for x in (type(expected_received_value), type(signals.value))) :
        # Testing the actual value sent to the data channel and the expected value using a tolerance due to floating
        # point arithmetic
        assert abs(signals.value - expected_received_value) < 0.00001


@pytest.mark.parametrize("new_write_access, is_channel_connected, tooltip, is_app_read_only", [
    (True, True, "Write Access and Connected Channel", False),
    (False, True, "Only Connected Channel", False),
    (True, False, "Only Write Access", False),
    (False, False, "No Write Access and No Connected Channel", False),

    (True, True, "Write Access and Connected Channel", True),
    (False, True, "Only Connected Channel", True),
    (True, False, "Only Write Access", True),
    (False, False, "No Write Access and No Connected Channel", True),

    (True, True, "", False),
    (True, True, "", True),
])
def test_write_access_changed(qapp, qtbot, signals, new_write_access, is_channel_connected, tooltip, is_app_read_only):
    """
    Test the widget's write access status and tooltip, which depends on the connection status of the data channel and
    the app's read-only status.

    Expectations:

    1. If the write access is set to False, the widget is read-only.
    2. If the data channel is disconnected, the widget's tooltip will  "PV is disconnected"
    3. If the data channel is connected, but it has no write access:
        a. If the app is read-only, the tooltip will read  "Running PyDM on Read-Only mode."
        b. If the app is not read-only, the tooltip will read "Access denied by Channel Access Security."

    Parameters
    ----------
    qapp : fixture
        pytest-qt qapp instance.
    qtbot : fixture
        pytest-qt window for widget testing
    signals : fixture
        The signals fixture, which provides access signals to be bound to the appropriate slots
    new_write_access : bool
        True if the widget has write access; False otherwise
    is_channel_connected : bool
        True if the data channel is connected; False otherwise
    tooltip : str
        The widget's tooltip
    is_app_read_only : bool
        True if the app is read-only; False otherwise
    """
    pydm_lineedit = PyDMLineEdit()
    qtbot.addWidget(pydm_lineedit)
    pydm_lineedit.show()

    set_read_only(is_app_read_only)

    pydm_lineedit.channel = "ca://MTEST"
    pydm_lineedit.setToolTip(tooltip)

    pydm_lineedit.connectionStateChanged(is_channel_connected)
    pydm_lineedit.writeAccessChanged(new_write_access)
    qapp.processEvents()

    if new_write_access:
        assert not pydm_lineedit.isReadOnly()
    else:
        assert pydm_lineedit.isReadOnly()

    actual_tooltip = pydm_lineedit.toolTip()
    if not pydm_lineedit._connected:
        assert "PV is disconnected." in actual_tooltip
    elif not new_write_access:
        if is_app_read_only:
            assert "Running PyDM on Read-Only mode." in actual_tooltip
        else:
            assert "Access denied by Channel Access Security." in actual_tooltip


@pytest.mark.parametrize("is_precision_from_pv, pv_precision, non_pv_precision", [
    (True, 1, 3),
    (False, 5, 3),
    (True, 6, 0),
    (True, 3, None),
    (False, 3, None),
])
def test_precision_change(qtbot, signals, is_precision_from_pv, pv_precision, non_pv_precision):
    """
    Test setting the precision for the widget's value.

    Expectations:

    1. If the precision comes from the PV, emit a prec_signal to change the widget's precision.
    2. If this is a non-PV precision, set the widget's precision directly.

    Parameters
    ----------
    qtbot : fixture
        pytest-qt window for widget testing
    signals : fixture
        The signals fixture, which provides access signals to be bound to the appropriate slots
    is_precision_from_pv : bool
        True if the precision value comes from the PV; False otherwise
    pv_precision : int
        The PV precision value to set to the widget
    non_pv_precision : int
        The non-PV precision value to set to the widget
    """
    pydm_lineedit = PyDMLineEdit()
    qtbot.addWidget(pydm_lineedit)

    pydm_lineedit.precisionFromPV = is_precision_from_pv
    pydm_lineedit.precision = non_pv_precision

    if is_precision_from_pv:
        signals.prec_signal[type(pv_precision)].connect(pydm_lineedit.precisionChanged)
        signals.prec_signal.emit(pv_precision)

        assert pydm_lineedit.precision == pv_precision
    else:
        assert pydm_lineedit.precision == (non_pv_precision if non_pv_precision is not None else 0)


@pytest.mark.parametrize("new_unit", [
    "s",
    "light years",
    "",
])
def test_unit_change(qtbot, signals, new_unit):
    """
    Test setting the widget's unit.

    Parameters
    ----------
    qtbot : fixture
        pytest-qt window for widget testing
    signals : fixture
        The signals fixture, which provides access signals to be bound to the appropriate slots
    new_unit : str
        The new unit to set to the widget
    """
    pydm_lineedit = PyDMLineEdit()
    qtbot.addWidget(pydm_lineedit)

    signals.unit_signal[str].connect(pydm_lineedit.unitChanged)
    signals.unit_signal.emit(new_unit)

    assert pydm_lineedit._unit == new_unit


def find_action_from_menu(menu, action_name):
    """
    Verify if an action (a conversion unit) is available in a context menu.

    Parameters
    ----------
    menu : QMenu
        The context menu of a widget
    action_name : str
        A menu text item

    Returns
    -------
    True if the action name is found in the menu; False otherwise
    """
    for action in menu.actions():
        if action.menu():
            # The action will always contain a menu, so the status will be created
            status = find_action_from_menu(action.menu(), action_name)
        if not action.isSeparator():
            if action_name == action.text():
                return True
    return status


@pytest.mark.parametrize("unit, show_units", [
    ("ms", True),
    ("s", False),
    ("MHz", True),
    ("V", True),
    ("in", True),
    ("mrad", True),
    ("MA", True),
])
def test_create_unit_options(qtbot, unit, show_units):
    """
    Test to ensure the context menu contains all applicable units that can be converted to from a given unit.

    Expectations:

    Given a unit and the show unit flag being set to True, the context menu must contain a unit conversion menu that
    has all the applicable units the original unit can be converted to.

    Parameters
    ----------
    qtbot : fixture
        pytest-qt window for widget testing
    unit : str
        The original unit
    show_units : bool
        If True, the context menu must contain all the conversion units; False otherwise
    """
    pydm_lineedit = PyDMLineEdit()
    qtbot.addWidget(pydm_lineedit)

    pydm_lineedit._unit = unit
    pydm_lineedit._show_units = show_units
    pydm_lineedit.create_unit_options()

    menu = pydm_lineedit.widget_ctx_menu()
    action_menu = menu.menuAction().menu()

    if unit and show_units:
        # Get all the units that can be converted to from the original unit
        units = find_unit_options(pydm_lineedit._unit)
        for unit in units:
            # For each of such unit, make sure it's in the context menu
            assert find_action_from_menu(action_menu, unit)
    else:
        # If no unit is provided, or the unit isn't required to be shown, the context menu should display
        # "No Unit Conversions found"
        assert find_action_from_menu(action_menu, "No Unit Conversions found")


@pytest.mark.parametrize("value, precision, unit, show_unit, expected_format_string", [
    (123, 0, "s", True, "{:.0f} s"),
    (123.456, 3, "mV", True, "{:.3f} mV"),
])
def test_apply_conversion(qtbot, value, precision, unit, show_unit, expected_format_string):
    """
    Test the unit conversion by examining the resulted format string.

    Expectations:

    Provided with the value, precision, unit, and the show unit Boolean flag by the user, this function must provide
    the correct format string to format the displayed value for the widget.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    value : int, float, bin, hex, numpy.array
        The value to be converted
    precision : int
        The
    unit : str
        The unit of the new value
    show_units : bool
        True if the value unit is to be displayed. False otherwise
    expected_format_string : str
        The expected format string that will produce the correct displayed value after the conversion
    """
    pydm_lineedit = PyDMLineEdit()
    qtbot.addWidget(pydm_lineedit)

    pydm_lineedit.value = value
    pydm_lineedit._unit = unit
    pydm_lineedit._prec = precision
    pydm_lineedit.showUnits = show_unit

    pydm_lineedit.apply_conversion(unit)
    assert pydm_lineedit.format_string == expected_format_string


@pytest.mark.parametrize("value, has_focus, channel_type, display_format, precision, scale, unit, show_units", [
    (123, True, int, DisplayFormat.Default, 3, 1, "s", True),
    (123, False, int, DisplayFormat.Default, 3, 1, "s", True),
    (123, True, int, DisplayFormat.Default, 3, 1, "s", False),
    (123, False, int, DisplayFormat.Default, 3, 1, "s", False),
    (123.45, True, float, DisplayFormat.Decimal, 3, 2, "m", True),
    (1e3, True, int, DisplayFormat.Exponential, 2, 2, "GHz", True),
    (0x1FF, True, int, DisplayFormat.Hex, 0, 1, "Me", True),
    (0b100, True, int, DisplayFormat.Binary, 0, 1, "KB", True),
    (np.array([123, 456]), str, True, DisplayFormat.Default, 3, 2, "degree", True),
])
def test_set_display(qtbot, qapp, value, has_focus, channel_type, display_format, precision, scale, unit, show_units):
    """
    Test the widget's displayed value.

    Expectations:

    The widget displays the value according to the display format from parsing the value and from the data type.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    signals : fixture
        The signals fixture, which provides access signals to be bound to the appropriate slots
    has_focus : Boolean
        True if the widget has the focus; False otherwise
    channel_type : type
        The type of the input value
    display_format : DisplayFormat
        The format to display the value
    precision : int
        The number of decimal places to consider for the new value
    scale : int
        The factor to reduce or magnify the new value
    unit : str
        The unit of the new value
    show_units : bool
        True if the value unit is to be displayed. False otherwise
    """
    pydm_lineedit = PyDMLineEdit()
    qtbot.addWidget(pydm_lineedit)
    with qtbot.waitExposed(pydm_lineedit):
        pydm_lineedit.show()

    pydm_lineedit.value = value
    pydm_lineedit._unit = unit
    pydm_lineedit.displayFormat = display_format
    pydm_lineedit._scale = scale
    pydm_lineedit.channeltype = channel_type
    pydm_lineedit.showUnits = show_units
    pydm_lineedit._prec = precision
    pydm_lineedit._display = "Empty"
    pydm_lineedit._connected = True
    pydm_lineedit._write_access = True
    pydm_lineedit.check_enable_state()

    if has_focus:
        def wait_focus():
            pydm_lineedit.setFocus()
            qapp.processEvents()
            return pydm_lineedit.hasFocus()

        qtbot.waitUntil(wait_focus, timeout=5000)

        pydm_lineedit.set_display()

        # If there's no focus on the widget, its display will not be updated
        assert pydm_lineedit._display == "Empty"
    else:
        pydm_lineedit.clearFocus()
        def wait_nofocus():
            pydm_lineedit.clearFocus()
            return not pydm_lineedit.hasFocus()

        qtbot.waitUntil(wait_nofocus, timeout=5000)
        pydm_lineedit.set_display()

        new_value = value
        if not isinstance(value, (str, np.ndarray)):
            new_value *= pydm_lineedit.channeltype(pydm_lineedit._scale)

        new_value = parse_value_for_display(value=new_value, precision=precision, display_format_type=display_format,
                                            widget=pydm_lineedit)

        expected_display = str(new_value)
        if display_format == DisplayFormat.Default and not isinstance(value, np.ndarray):
            if isinstance(value, (int, float)):
                expected_display = str(pydm_lineedit.format_string.format(value))
        elif pydm_lineedit.showUnits:
            expected_display += " {}".format(unit)

        assert pydm_lineedit._display == expected_display


@pytest.mark.parametrize("displayed_value, focus_reason, expected_focus", [
    (True, Qt.TabFocusReason, True),
    (True, Qt.ActiveWindowFocusReason, True),
    (False, Qt.TabFocusReason, False),
    (False, Qt.ActiveWindowFocusReason, False),
    (False, Qt.MouseFocusReason, True)])
def test_focus_in_event(qtbot, qapp, displayed_value, focus_reason, expected_focus):
    """
    Ensure that the line edit's focusInEvent() override works as expected. When the widget has not yet been connected
    long enough to have received a value, it should specifically reject tab focus and active window focus events.
    """
    # Create a PyDMLineEdit and show it
    pydm_lineedit = PyDMLineEdit()
    qtbot.addWidget(pydm_lineedit)
    pydm_lineedit._has_displayed_value_yet = displayed_value
    with qtbot.waitExposed(pydm_lineedit):
        pydm_lineedit.show()

    def wait_focus(focus_state, wait_on_clear):
        """ Verify the current focus state of the line edit """
        if focus_state:
            pydm_lineedit.setFocus(Qt.OtherFocusReason)
        elif wait_on_clear:
            pydm_lineedit.clearFocus()
        qapp.processEvents()
        return pydm_lineedit.hasFocus() == focus_state

    # First set the current focus status opposite of that which we expect, to ensure we are actually
    # measuring the correct end state
    if expected_focus:
        pydm_lineedit.clearFocus()
        qtbot.waitUntil(functools.partial(wait_focus, False, True), timeout=5000)
    else:
        pydm_lineedit.setFocus(Qt.OtherFocusReason)
        qtbot.waitUntil(functools.partial(wait_focus, True, False), timeout=5000)

    # Set the focus, verify the result is what we are expecting
    if expected_focus:
        pydm_lineedit.setFocus(focus_reason)
    else:
        event_to_send = QFocusEvent(QEvent.FocusIn, reason=focus_reason)
        pydm_lineedit.focusInEvent(event_to_send)
    qtbot.waitUntil(functools.partial(wait_focus, expected_focus, False), timeout=5000)


@pytest.mark.parametrize("display_value", [
    "123",
    "123.456",
    "",
])
def test_focus_out_event(qtbot, qapp, display_value):
    """
    Test the widget's value revert capability if the user doesn't commit the value change.

    Expectations:

    If the user types in some value to the widget, but then does not commit the change, but simply leaves the widget,
    i.e. generating a focusOut event, the widget's value will revert to the original one.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    qapp : fixture
        The current pytest-qt app
    display_value : str
        The new value entered by the user before leaving the widget without committing this value
    """
    pydm_lineedit = PyDMLineEdit()
    qtbot.addWidget(pydm_lineedit)
    with qtbot.waitExposed(pydm_lineedit):
        pydm_lineedit.show()

    pydm_lineedit._connected = True
    pydm_lineedit._write_access = True
    pydm_lineedit.check_enable_state()

    pydm_lineedit._display = display_value
    def wait_focus():
        pydm_lineedit.setFocus()
        qapp.processEvents()
        return pydm_lineedit.hasFocus()

    qtbot.waitUntil(wait_focus, timeout=5000)

    pydm_lineedit.setText("Canceled after the focusOut event")

    def wait_nofocus():
        pydm_lineedit.clearFocus()
        qapp.processEvents()
        return not pydm_lineedit.hasFocus()

    qtbot.waitUntil(wait_nofocus, timeout=5000)

    # Make sure the widget still retains the previously set value after the focusOut event
    assert pydm_lineedit.text() == display_value

# --------------------
# NEGATIVE TEST CASES
# --------------------

@pytest.mark.parametrize("value, precision, initial_unit, unit, show_unit, expected", [
    (123, 0, None, int, True,
     ("Warning: Attempting to convert PyDMLineEdit unit, but no initial units supplied",)),
    (123.456, 3, float, int, True,
     ("Warning: Attempting to convert PyDMLineEdit unit, but ", "'float'>' can not be converted to ", "'int'>'.")),
    (123.456, 3, float, "foo", True,
     ("Warning: Attempting to convert PyDMLineEdit unit, but ", "'float'>' can not be converted to 'foo'.")),
    ("123.456", 3, "light years", "light years", False,
     ("Warning: Attempting to convert PyDMLineEdit unit, but 'light years' can not be converted to 'light years'.",)),
])
def test_apply_conversion_wrong_unit(qtbot, caplog, value, precision, initial_unit, unit, show_unit, expected):
    """
    Test the unit conversion error logging.

    Expectations:

    Errors result from unit conversion problems must be logged correctly.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    caplog : fixture
        To capture the log messages
    value : int, float, bin, hex, numpy.array
        The signals fixture, which provides access signals to be bound to the appropriate slots
    precision : int
        The
    unit : str
        The unit of the new value
    show_units : bool
        True if the value unit is to be displayed. False otherwise
    expected : str
        The expected logged error statement
    """
    pydm_lineedit = PyDMLineEdit()
    qtbot.addWidget(pydm_lineedit)

    pydm_lineedit.value = value
    pydm_lineedit._unit = initial_unit
    pydm_lineedit._prec = precision
    pydm_lineedit.showUnits = show_unit

    pydm_lineedit.apply_conversion(unit)
    assert all(x in caplog.text for x in expected)


@pytest.mark.parametrize("init_value, user_typed_value, display_format, precision, scale, unit,"
                         "show_units, expected_errors", [
    (123, 345.678, DisplayFormat.Default, 0, 1, "s", True, ("Error trying to set data ", "with type ", "int'>")),
])
def test_send_value_neg(qtbot, caplog, signals, init_value, user_typed_value, display_format, precision, scale, unit,
                        show_units, expected_errors):
    """
    Test sending the value to the channel error logging.

    Expectations:

    Errors result from data sending problems must be logged correctly.

    Parameters
    ----------
    qtbot : fixture
        pytest-qt window for widget testing
    caplog : fixture
        To capture the log messages
    signals : fixture
        The signals fixture, which provides access signals to be bound to the appropriate slots
    init_value : int, float, hex, bin, str, numpy.array
        The initial value currently assigned to the widget
    user_typed_value : int, float, hex, bin, str, numpy.array
        The new value as provided (typed in) by the user
    display_format : DisplayFormat
        The display format of the current value
    precision : int
        The number of decimal places to consider for the new value
    scale : int
        The factor to reduce or magnify the new value
    unit : str
        The unit of the new value
    show_units : bool
        True if the value unit is to be displayed. False otherwise
    expected_errors : str
        The expected logged error statement
    """
    pydm_lineedit = PyDMLineEdit()
    qtbot.addWidget(pydm_lineedit)

    pydm_lineedit.value = init_value
    pydm_lineedit.channeltype = type(init_value)
    pydm_lineedit.displayFormat = display_format
    pydm_lineedit._prec = precision
    pydm_lineedit._scale = scale
    pydm_lineedit._unit = unit
    pydm_lineedit.showUnits = show_units

    signal_type = type(user_typed_value)
    if signal_type == np.ndarray:
        signal_type = str
    pydm_lineedit.send_value_signal[signal_type].connect(signals.receiveValue)
    pydm_lineedit.send_value_signal[signal_type].connect(pydm_lineedit.channelValueChanged)
    pydm_lineedit.send_value()

    for record in caplog.records:
        assert record.levelno == ERROR
    assert all(x in caplog.text for x in expected_errors)



import functools
import importlib
import importlib.util
import logging
import ntpath
import os
import platform
import shlex
import sys
import types
import uuid
import errno

from typing import List, Optional

from qtpy import QtCore, QtGui, QtWidgets

from . import colors, macro, shortcuts
from .connection import close_widget_connections, establish_widget_connections
from .iconfont import IconFont
from .remove_protocol import protocol_and_address, remove_protocol, parsed_address
from .units import convert, find_unit_options, find_unittype

logger = logging.getLogger(__name__)


def is_ssh_session():
    """
    Whether or not this is a SSH session.

    Returns
    -------
    bool
        True if it is a ssh session, False otherwise.
    """
    return os.getenv('SSH_CONNECTION') is not None


def setup_renderer():
    """
    This utility function reverts the renderer to Software rendering if it is
    running in a SSH session.
    """
    if is_ssh_session():
        logger.info('Using PyDM via SSH. Reverting to Software Rendering.')
        from qtpy.QtCore import QCoreApplication, Qt
        from qtpy.QtQuick import QQuickWindow, QSGRendererInterface

        QCoreApplication.setAttribute(Qt.AA_UseSoftwareOpenGL)
        QQuickWindow.setSceneGraphBackend(QSGRendererInterface.Software)


def is_pydm_app(app=None):
    """
    Check whether or not `QApplication.instance()` is a PyDMApplication.

    Parameters
    ----------
    app : QApplication, Optional
        The app to inspect. If no application is provided the current running `QApplication` will be queried.

    Returns
    -------
    bool
        True if it is a PyDMApplication, False otherwise.
    """
    from qtpy.QtWidgets import QApplication

    from ..application import PyDMApplication
    if app is None:
        app = QApplication.instance()
    if isinstance(app, PyDMApplication):
        return True
    else:
        return False


def is_qt_designer():
    """
    Check whether or not running inside Qt Designer.

    Returns
    -------
    bool
        True if inside Designer, False otherwise.
    """
    from ..qtdesigner import DesignerHooks
    return DesignerHooks().form_editor is not None


def get_designer_current_path():
    """
    Fetch the absolute path for the current active form at Qt Designer.

    Returns
    -------
    path : str, None
        The absolute path for the current active form or None in case not
        available
    """
    if not is_qt_designer():
        return None

    from ..qtdesigner import DesignerHooks
    form_editor = DesignerHooks().form_editor
    win_manager = form_editor.formWindowManager()
    form_window = win_manager.activeFormWindow()
    if form_window is None and win_manager.formWindowCount() > 0:
        form_window = win_manager.formWindow(0)
    if form_window is not None:
        abs_dir = form_window.absoluteDir()
        if abs_dir:
            return abs_dir.absolutePath()

    return None


def path_info(path_str):
    """
    Retrieve basic information about the given path.

    Parameters
    ----------
    path_str : str
        The path from which to extract information.

    Returns
    -------
    tuple
        base dir, file name, list of args
    """
    if platform.system() == "Windows":
        os_path_mod = ntpath
    else:
        os_path_mod = os.path

    dir_name, other_parts = os_path_mod.split(path_str)
    split = shlex.split(other_parts)
    file_name = split.pop(0)
    args = split

    return dir_name, file_name, args


def _extensions(fname):
    name = os.path.basename(fname)
    MAX_ITER = 10
    exts = []
    for i in range(MAX_ITER):
        new_name, ext = os.path.splitext(name)
        if ext:
            exts.insert(0, ext)
        if name == new_name:
            break
        name = new_name
    return exts


def _screen_file_extensions(preferred_extension):
    """
    Return a prioritized list of screen file extensions.

    Include .ui & .py files (also .adl files if adl2pydm installed).
    Prefer extension as described by fname.
    """
    extensions = [".py", ".ui"]  # search for screens with these extensions
    try:
        import adl2pydm  # proceed only if package is importable  # noqa: F401
        extensions.append(".adl")
    except ImportError:
        pass

    # don't search twice for preferred extension
    if preferred_extension in extensions:
        extensions.remove(preferred_extension)

    # search first for preferred extension
    extensions.insert(0, preferred_extension)
    return extensions


def find_file(fname, base_path=None, mode=None, extra_path=None, raise_if_not_found=False):
    """
    Look for files at the search paths common to PyDM.

    The search order is as follows:

    * The ``base_path`` argument
    * Qt Designer Path - the path for the current form as reported by the
      designer
    * The current working directory
    * Directories listed in ``extra_path``
    * Directories listed in the environment variable ``PYDM_DISPLAYS_PATH``

    Parameters
    ----------
    fname : str
        The file name. Environment variables, ~ and ~user constructs before
        search.
    base_path : str
        The directory name of a file pathname from a display, if any
    mode : int
        The mode required for the file, defaults to os.F_OK | os.R_OK.
        Which ensure that the file exists and we can read it.
    extra_path : list
        Additional paths to look for file.
    raise_if_not_found : bool 
        Flag which if False will add a check that raises a FileNotFoundError 
        instead of returning None when the file is not found. 

    Returns
    -------
    file_path : str
        Returns the file path or None in case the file was not found
    """
    fname = os.path.expanduser(os.path.expandvars(fname))

    if mode is None:
        mode = os.F_OK | os.R_OK

    x_path = []

    if base_path:
        x_path.extend([os.path.abspath(base_path)])

    if is_qt_designer():
        designer_path = get_designer_current_path()
        if designer_path:
            x_path.extend([designer_path])

    # Current working directory
    x_path.extend([os.getcwd()])
    if extra_path:
        if not isinstance(extra_path, (list, tuple)):
            extra_path = [extra_path]
        extra_path = [os.path.expanduser(os.path.expandvars(x)) for x in extra_path]
        x_path.extend(extra_path)

    pydm_search_path = os.getenv("PYDM_DISPLAYS_PATH", None)
    if pydm_search_path:
        x_path.extend(pydm_search_path.split(os.pathsep))

    for idx, path in enumerate(x_path):
        x_path[idx] = os.path.expanduser(os.path.expandvars(path))

    root, ext = os.path.splitext(fname)

    # loop through the possible screen file extensions
    for e in _screen_file_extensions(ext):
        file_path = which(str(root) + str(e), mode=mode, pathext=e, extra_path=x_path)
        if file_path is not None:
            break  # pick the first screen file found
    
    if raise_if_not_found:
        if not file_path:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fname)

    return file_path


def find_display_in_path(file, mode=None, path=None, pathext=None):
    """
    Look for a display file in a given path.
    This is basically a wrapper on top of the ``which``
    command defined below so we don't need to keep redefining
    the ``PYDM_DISPLAYS_PATH`` variable.

    Parameters
    ----------
    file : str
        The file name.
    mode : int
        The mode required for the file, defaults to os.F_OK | os.R_OK.
        Which ensure that the file exists and we can read it.
    path : str
        The PATH string.

    Returns
    -------
    str
        Returns the full path to the file or None in case it was not found.
    """
    if pathext is None and sys.platform == "win32":
        pathext = ".ui"
    if path is None:
        path = os.getenv("PYDM_DISPLAYS_PATH", None)
    if mode is None:
        mode = os.F_OK | os.R_OK

    return which(file, mode, path, pathext=pathext)


def which(cmd, mode=os.F_OK | os.X_OK, path=None, pathext=None, extra_path=None):
    """Given a command, mode, and a PATH string, return the path which
    conforms to the given mode on the PATH, or None if there is no such
    file.
    `mode` defaults to os.F_OK | os.X_OK. `path` defaults to the result
    of os.environ.get("PATH"), or can be overridden with a custom search
    path.
    Note: This function was backported from the Python 3 source code and modified
    to deal with the case in which we WANT to look at the path even with a relative
    path.
    """

    # Check that a given file can be accessed with the correct mode.
    # Additionally check that `file` is not a directory, as on Windows
    # directories pass the os.access check.
    def _access_check(fn, mode):
        return (os.path.exists(fn) and os.access(fn, mode) and
                not os.path.isdir(fn))

    # If we're given a path with a directory part, look it up directly
    # rather than referring to PATH directories. This includes checking
    # relative to the current directory, e.g. ./script
    # if os.path.dirname(cmd):
    #     if _access_check(cmd, mode):
    #         return cmd
    #     return None

    if path is None:
        path = os.environ.get("PATH", os.defpath)
    if not path:
        return None
    path = path.split(os.pathsep)

    if extra_path is not None:
        path = extra_path + path

    if sys.platform == "win32":
        # The current directory takes precedence on Windows.
        if os.curdir not in path:
            path.insert(0, os.curdir)

        # PATHEXT is necessary to check on Windows.
        if pathext is None:
            pathext = os.environ.get("PATHEXT", "")
        pathext = pathext.split(os.pathsep)
        # See if the given file matches any of the expected path
        # extensions. This will allow us to short circuit when given
        # "python.exe". If it does match, only test that one, otherwise we
        # have to try others.
        if any(cmd.lower().endswith(ext.lower()) for ext in pathext):
            files = [cmd]
        else:
            files = [cmd + ext for ext in pathext]
    else:
        # On other platforms you don't have things like PATHEXT to tell you
        # what file suffixes are executable, so just pass on cmd as-is.
        files = [cmd]

    seen = set()
    for dir_ in path:
        normdir = os.path.normcase(dir_)
        if normdir not in seen:
            seen.add(normdir)
            for thefile in files:
                name = os.path.join(dir_, thefile)
                if _access_check(name, mode):
                    return name
    return None


def only_main_thread(func):
    """
    Decorator that wraps a function which should only be executed at the Qt
    main thread.

    The decorator will log an error message and raise a RuntimeError if the
    decorated function is invoked from a thread other than the Qt main one.

    Parameters
    ----------
    func : callable
        The function to wrap

    Returns
    -------
    wrapper
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        main_t = QtWidgets.QApplication.instance().thread()
        curr_t = QtCore.QThread.currentThread()
        if curr_t != main_t:
            msg = "{}.{} can only be invoked from the main Qt thread.".format(
                func.__module__, func.__name__
            )
            logger.error(msg)
            raise RuntimeError(msg)
        return func(*args, **kwargs)

    if not callable(func):
        raise ValueError("Parameter must be a callable.")

    return wrapper


def log_failures(
    logger: logging.Logger,
    explanation: str = "Failed to run {func.__name__}",
    include_traceback: bool = False,
    level: int = logging.WARNING,
):
    """
    Decorator that wraps a function to be run.

    Exceptions raised while executing that function will be logged.
    In case of an exception, the wrapper will return ``None``.

    Parameters
    ----------
    logger : logging.Logger
        The logger instance to log messages to.
    explanation : str, optional
        The explanation message to include.  Format arguments include:
            ``func``, ``args``, ``kwargs``, and the exception ``ex``.
    include_traceback : bool, optional
        Include traceback information in the log message.
    level : int, optional
        Logging level to use.
    """
    def wrapper(func: callable):
        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as ex:
                msg = explanation.format(
                    func=func, args=args, kwargs=kwargs, ex=ex
                )
                if include_traceback:
                    logger.log(level, msg, exc_info=ex)
                else:
                    logger.log(level, msg)
                return None

        return wrapped

    return wrapper


def import_module_by_filename(
    source_filename: str, *,
    add_to_modules: bool = True
) -> types.ModuleType:
    """
    For a given source filename, import it and search for objects.

    Parameters
    ----------
    source_filename : str
        The source code filename.

    add_to_modules : bool, optional, keyword-only
        Add the imported module to ``sys.modules``.  Defaults to ``True``.

    Returns
    -------
    module : types.ModuleType
        The imported module.
    """
    module_dir = os.path.dirname(os.path.abspath(source_filename))
    if module_dir not in sys.path:
        sys.path.append(module_dir)

    module_name = str(uuid.uuid4())
    spec = importlib.util.spec_from_file_location(module_name, source_filename)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    if add_to_modules:
        sys.modules[module_name] = module
    return module

def get_clipboard() -> Optional[QtGui.QClipboard]:
    """Get the clipboard instance. Requires a QApplication."""
    app = QtWidgets.QApplication.instance()
    if app is None:
        return None

    return QtWidgets.QApplication.clipboard()


def get_clipboard_modes() -> List[int]:
    """
    Get the clipboard modes for the current platform.

    Returns
    -------
    list of int
        Qt-specific modes to try for interacting with the clipboard.
    """
    clipboard = get_clipboard()
    if clipboard is None:
        return []

    if platform.system() == "Linux":
        # Mode selection is only valid for X11.
        return [
            QtGui.QClipboard.Selection,
            QtGui.QClipboard.Clipboard
        ]

    return [QtGui.QClipboard.Clipboard]


def copy_to_clipboard(text: str, *, quiet: bool = False):
    """
    Copy ``text`` to the clipboard.

    Parameters
    ----------
    text : str
        The text to copy to the clipboard.

    quiet : bool, optional, keyword-only
        If quiet is set, do not log the copied text.  Defaults to False.
    """
    clipboard = get_clipboard()
    if clipboard is None:
        return None

    for mode in get_clipboard_modes():
        clipboard.setText(text, mode=mode)
        event = QtCore.QEvent(QtCore.QEvent.Clipboard)
        app = QtWidgets.QApplication.instance()
        if app is not None:
            app.sendEvent(clipboard, event)

    if not quiet:
        logger.warning(
            (
                "Copied text to clipboard:\n"
                "-------------------------\n"
                "%s\n"
                "-------------------------\n"
            ),
            text
        )


def get_clipboard_text() -> str:
    """
    Get ``text`` from the clipboard. If unavailable or unset, empty string.

    Returns
    -------
    str
        The clipboard text, if available.
    """
    clipboard = get_clipboard()
    if clipboard is None:
        return ""
    for mode in get_clipboard_modes():
        text = clipboard.text(mode=mode)
        if text:
            return text
    return ""



import pytest
from ...utilities import iconfont
from qtpy import QtGui, QtCore


def test_icon_font_constructor(qtbot):
    icon_f = iconfont.IconFont()
    icon_f2 = iconfont.IconFont()
    assert (icon_f is icon_f2)


def test_icon_font_load_font(qtbot):
    icon_f = iconfont.IconFont()
    with pytest.raises(OSError):
        icon_f.char_map = None
        icon_f.load_font('foo', icon_f.charmap_file)
    with pytest.raises(OSError):
        icon_f.char_map = None
        icon_f.load_font(icon_f.charmap_file, 'foo')
    icon_f.load_font(icon_f.font_file, icon_f.charmap_file)
    assert (icon_f.char_map is not None)


def test_icon_font_get_char_for_name(qtbot):
    icon_f = iconfont.IconFont()
    c = icon_f.get_char_for_name('cogs')
    assert (c == u'\uf085')

    with pytest.raises(ValueError):
        icon_f.get_char_for_name('foo')


def test_icon_font_font(qtbot):
    icon_f = iconfont.IconFont()
    f = icon_f.font(12)
    assert(f.family() == icon_f.font_name)
    assert(f.pixelSize() == 12)


def test_icon_font_icon(qtbot):
    icon_f = iconfont.IconFont()
    ico = icon_f.icon('cogs', color=None)
    ico1 = icon_f.icon('cogs', color=QtGui.QColor(255, 0, 0))
    with pytest.raises(ValueError):
        ico_invalid = icon_f.icon('foo', color=None)


def test_char_icon_engine(qtbot):
    engine = iconfont.CharIconEngine(iconfont.IconFont(), 'cogs', color=None)
    pm = engine.pixmap(QtCore.QSize(32, 32), mode=QtGui.QIcon.Normal, state=QtGui.QIcon.On)
    pm = engine.pixmap(QtCore.QSize(32, 32), mode=QtGui.QIcon.Disabled, state=QtGui.QIcon.On)